from __future__ import absolute_import
from __future__ import print_function
import os
import logging
import six.moves.configparser
from collections import OrderedDict
from scipy.interpolate import interp1d
from bisect import bisect_left
import numpy as np
from itertools import tee

import json
import matplotlib.pyplot as plt
import matplotlib.ticker as plticker
import matplotlib.gridspec as gridspec
from matplotlib import rcParams
from matplotlib.axes import Axes
#import pandas as pd
# SEABORN creates by default plots with a filled background!!
#import seaborn as sns
from astropy.io import ascii
from astropy.io import fits

import sys
import pyp_beagle.dependencies.WeightedKDE 
from pyp_beagle.dependencies.walker_random_sampling import WalkerRandomSampling
import pyp_beagle.dependencies.autoscale
from pyp_beagle.dependencies import FillBetweenStep
import pyp_beagle.dependencies.set_shared_labels  as shLab

from .beagle_utils import BeagleDirectories, prepare_plot_saving, set_plot_ticks, plot_exists
from .beagle_filters import PhotometricFilters
from .beagle_summary_catalogue import BeagleSummaryCatalogue
#from beagle_residual_photometry import ResidualPhotometry
from .beagle_multinest_catalogue import MultiNestCatalogue
from .beagle_posterior_predictive_checks import PosteriorPredictiveChecks
from .beagle_mock_catalogue import BeagleMockCatalogue

# See here
# http://peak.telecommunity.com/DevCenter/PythonEggs#accessing-package-resources
# for an explanation on this approach to include data files
from pkg_resources import resource_stream
import six
from six.moves import zip
from six.moves import range

TOKEN_SEP = ":"
microJy = np.float32(1.E-23 * 1.E-06)
nanoJy = np.float32(1.E-23 * 1.E-09)

p_value_lim = 0.05


def pairwise(iterable):
    "s -> (s0,s1), (s1,s2), (s2, s3), ..."
    a, b = tee(iterable)
    next(b, None)
    return zip(a, b)

class ObservedSpectrum(object):

    def __init__(self):

        self.description  = None

    def configure(self, param_file=None, config=None):

        if param_file is None:
            param_file = os.path.join(BeagleDirectories.results_dir, BeagleDirectories.param_file)

        if config is None:
            config = six.moves.configparser.SafeConfigParser()
            config.read(param_file)

        line = config.get('main', 'SPECTRUM FILE DESCRIPTION')

        self.description = { 
                "wl"        : {"colName" : None, "conversion" : None, "dispersion" : None, "type" : None}, 
                "flux"      : {"colName" : None, "conversion" : None}, 
                "fluxerr"   : {"colName" : None},  
                "sky"       : {"colName" : None},  
                "mask"      : {"colName" : None},  
                "redshift"  : {"keyword" : "redshift"},  
                "min_rel_err" : None,  
                }

        for key, value in six.iteritems(self.description):

            token = key + TOKEN_SEP

            if isinstance(value, dict):
                for in_key, in_value in six.iteritems(value):
                    in_token = token + in_key + TOKEN_SEP
                    if in_token in line:
                        self.description[key][in_key] = line.split(in_token)[1].split(" ")[0]

            else:
                if token in line:
                    self.description[key] = line.split(token)[1].split(" ")[0]
                

    def load(self, file_name):

        """ 
        Load an observed spectrum. It automatically
        detects, and loads, FITS or ASCII files depending on the suffix.

        Parameters
        ----------
        file_name : str
            Contains the file name of the spectrum.
        """

        if self.description is None:
            msg = ("The `ObservedSpectrum.description` must be set through "
                    "the `configure` method!")
            raise AttributeError(msg)


        hdu = fits.open(os.path.expandvars(file_name))
        data = hdu[1].data

        self.data = dict()

        # Set the array containing the wavelength bins of the spectrum 
        try:
            self.data['wl'] = np.array(data[self.description["wl"]["colName"]])
        except:
            msg = ("`" + self.description["wl"]["colName"] + "` not found "
                    "in the current spectrum!")
            raise AttributeError(msg)

        if self.description["wl"]["conversion"] is not None:
            self.data['wl'] *= np.float(self.description["wl"]["conversion"])

        # Set the array containing the flux
        try:
            self.data['flux'] = np.array(data[self.description["flux"]["colName"]])
        except:
            msg = ("`" + self.description["flux"]["colName"] + "` not found "
                    "in the current spectrum!")
            raise AttributeError(msg)

        if self.description["flux"]["conversion"] is not None:
            self.data['flux'] *= np.float(self.description["flux"]["conversion"])

        # Set the array containing the flux error
        if self.description["fluxerr"]["colName"] is not None:
            try:
                self.data['fluxerr'] = np.array(data[self.description["fluxerr"]["colName"]])
            except:
                msg = ("`" + self.description["fluxerr"]["colName"] + "` not found "
                        "in the current spectrum!")
                raise AttributeError(msg)

        # Set the redshift
        self.data['redshift'] = None
        if self.description["redshift"]["keyword"] is not None:
            try:
                self.data['redshift'] = np.float(hdu[1].header[self.description["redshift"]["keyword"]])
            except:
                pass

        # Set the mask spectrum
        if self.description["mask"]["colName"] is not None:
            self.data['mask'] = np.array(data[self.description["mask"]["colName"]], dtype=bool)

        hdu.close()

class Spectrum(object):

    def __init__(self, params_file,
            **kwargs):

        self.inset_fontsize = BeagleDirectories.inset_fontsize_fraction * BeagleDirectories.fontsize

        self.observed_spectrum = ObservedSpectrum()

        self.multinest_catalogue = MultiNestCatalogue()

        self.mock_catalogue = kwargs.get('mock_catalogue')

        #self.residual = ResidualPhotometry()

        self.PPC = PosteriorPredictiveChecks()

        if kwargs.get('line_labels_json') is None:
            self.line_labels = json.load(resource_stream(__name__, 'files/emission_lines.json'), 
                    object_pairs_hook=OrderedDict)
        else:
            with open(kwargs.get('line_labels_json')) as f:    
                 self.line_labels = json.load(f, object_pairs_hook=OrderedDict)

        self.plot_line_labels = kwargs.get('plot_line_labels', False)

        self.resolution = kwargs.get('resolution')
    
        self.wl_range = kwargs.get('wl_range')

        self.wl_units = kwargs.get('wl_units', 'micron')

        self.wl_rest = kwargs.get('wl_rest', False)

        self.log_flux = kwargs.get('plot_log_flux', False)

        self.plot_full_SED = kwargs.get('plot_full_SED', False)

        self.show_residual = kwargs.get('show_residual', False)

        self.print_ID = kwargs.get('print_ID', False)

        self.draw_steps = kwargs.get('draw_steps', False)

        self.n_SED_to_plot = kwargs.get('n_SED_to_plot', 100)

        self.plot_suffix = kwargs.get('plot_suffix')

    def plot_marginal(self, ID, 
            observation_name=None,
            max_interval=95.0,
            print_text=False, 
            replot=False):    
        """ 
        Plot the fluxes predicted by BEAGLE.

        The fluxes here considered are those predicted by BEAGLE, given the
        posterior distribution of the model parameters. These are *not*
        replicated data.

        Parameters
        ----------
        ID : int
            ID of the galaxy whose marginal photometry will be plotted.

        max_interval : float, optional
            The marginal photometry is shown to include `max_interval`
            probability, e.g. `max_interval` = 68. will show the 68 % (i.e.
            '1-sigma') (central) credible region of the marginal photometry.

        print_text : bool, optional
            Whether to print further information on the plot, such as
            chi-square, p-value, or leave it empty and neat.

        print_text : bool, optional
            Whether to print the object ID on the top of the plot.
        """

        # Factor to convert angstrom to input units
        if self.wl_units == 'micron':
            wl_factor = 1.E+04
            xlabel = "\mu\\textnormal{m}"
        elif self.wl_units == 'nm':
            wl_factor = 1.E+01
            xlabel = "\\textnormal{nm}"
        elif self.wl_units == 'ang':
            wl_factor = 1.
            xlabel = "\\textnormal{\\AA}"
        else:
            raise ValueError("Wavelength units `" + self.wl_units + "` not recognised!")

        # If needed load the observed spectrum
        if observation_name is not None:
            self.observed_spectrum.load(observation_name)
    
        # Name of the output plot
        plot_suffix = '' 
        if self.plot_suffix is not None: plot_suffix = '_' + self.plot_suffix
        if self.wl_rest:
            plot_name = str(ID) + '_BEAGLE_marginal_SED_spec_rest_wl' + plot_suffix + '.pdf'
        else:
            plot_name = str(ID) + '_BEAGLE_marginal_SED_spec' + plot_suffix + '.pdf'

        # Check if the plot already exists
        if plot_exists(plot_name) and not replot:
            logging.warning('The plot "' + plot_name + '" already exists. \n Exiting the function.')
            return

        # The observed spectrum
        observation = self.observed_spectrum

        # Add to the error array the minimum relative error thet BEAGLE allows
        # one to add to the errors quoted in the catalogue
        #for i, err in enumerate(self.filters.data['flux_errcolName']):
        #    tmp_err = observation[0][err]
        #    if tmp_err > 0.:
        #        obs_flux_err[i] = observation[0][err]*aper_corr*self.filters.units / nanoJy
        #        obs_flux_err[i] = (np.sqrt( (obs_flux_err[i]/obs_flux[i])**2 +
        #                np.float32(self.filters.data['min_rel_err'][i])**2) *
        #                obs_flux[i])
        #    else:
        #        obs_flux_err[i] = tmp_err

        #ok = np.where(obs_flux_err > 0.)[0]

        # Open the file containing BEAGLE results
        fits_file = os.path.join(BeagleDirectories.results_dir,
                str(ID) + '_' + BeagleDirectories.suffix + '.fits.gz')

        hdulist = fits.open(fits_file)

        # Read the template wl array, and the 2D flux array
        model_wl = hdulist['marginal sed wl'].data['wl'][0,:]
        model_fluxes = hdulist['marginal sed'].data

        # Read the posterior probability
        probability = hdulist['posterior pdf'].data['probability']

        # Now compute for each wl bin the sorted fluxes, you will need this to
        # calculate the median and percentiles for each wl bin
        sort_indices = np.argsort(model_fluxes, axis=0)

        # Now it's time to compute the median (observed-frame) SED and its percentiles
        n_wl = model_fluxes.shape[1]

        median_flux = np.zeros(n_wl)
        lower_flux = np.zeros(n_wl)
        upper_flux = np.zeros(n_wl)

        for i in range(n_wl):

            # Compute the cumulative probability
            # ******************************************************************
            # Here you must simply use `cumsum`, and not `cumtrapz` as in
            # beagle_utils.prepare_violin_plot, since the output of MultiNest are a set
            # of weights (which sum up to 1) associated to each set of parameters (the
            # `p_j` of equation 9 of Feroz+2009), and not a probability density (as the
            # MultiNest README would suggest).
            # ******************************************************************
            sort_ = sort_indices[:,i]
            cumul_pdf = np.cumsum(probability[sort_])
            cumul_pdf /= cumul_pdf[len(cumul_pdf)-1]
            
            # Get the interpolant of the cumulative probability
            f_interp = interp1d(cumul_pdf, model_fluxes[sort_,i])

            # The median corresponds to a cumulative probability = 0.5
            median_flux[i] = f_interp(0.5)

            # Compute the percentiles for the different credible regions
            lev = (1.-max_interval/100.)/2.
            lower_flux[i] = f_interp(lev)

            lev = 1.-(1.-max_interval/100.)/2.
            upper_flux[i] = f_interp(lev)
    
        # Set the plot limits from the minimum and maximum wl_eff
        axs = list()
        residual_axs = list()
        if observation.data['redshift'] is not None:
            redshift = observation.data['redshift']
        else:
            _redshifts =  hdulist['galaxy properties'].data['redshift']
            _, _counts = np.unique(_redshifts, return_counts=True)
            if len(_counts) > 1:
                raise ValueError("The `redshift` of the object is not unique!")
            redshift = _redshifts[0]

        z1 = 1. + redshift

        if self.wl_rest:
            data_wl = observation.data['wl'] / z1
            data_flux = observation.data['flux'] * z1
            data_flux_err = observation.data['fluxerr'] * z1

            model_wl /= z1
            median_flux *= z1
            lower_flux *= z1
            upper_flux *= z1
        else:
            data_wl = observation.data['wl']
            data_flux = observation.data['flux']
            data_flux_err = observation.data['fluxerr']

        # Convert to correct wl units
        model_wl /= wl_factor
        data_wl /= wl_factor

        # Load the mask spectrum if exists
        mask_color = "grey"
        has_mask = False
        data_mask = np.ones(len(data_wl), dtype=bool)
        if "mask" in observation.data:
            data_mask = observation.data['mask']
            has_mask = True

        # Create masked versions of arrays
        slices = list()
        masked_array = np.ma.array(data_wl, mask=~data_mask)
        slices.append(np.ma.clump_unmasked(masked_array))
        slices.append(np.ma.clump_masked(masked_array))

        model_mask = np.ones(len(model_wl), dtype=bool)
        if "marginal sed mask" in hdulist:
            model_mask = np.array(hdulist['marginal sed mask'].data['mask'][0], dtype=bool)

        # Create masked versions of arrays
        slices_model = list()
        masked_array = np.ma.array(model_wl, mask=~model_mask)
        slices_model.append(np.ma.clump_unmasked(masked_array))
        slices_model.append(np.ma.clump_masked(masked_array))


        if self.show_residual:
            n_outer = 2
        else:
            n_outer = 1

        if self.wl_range is None:
            n_ranges = 1
        else:
            if len(self.wl_range) == 2:
                n_ranges = 1
            else:
                n_ranges = int(1.*len(self.wl_range)/2.)

        fig = plt.figure(figsize=(12,8))
        if self.show_residual:
            fig, axs_ = plt.subplots(n_outer, n_ranges, gridspec_kw = {'height_ratios':[3, 1]})
        else:
            fig, axs_ = plt.subplots(n_outer, n_ranges)

        fig.subplots_adjust(wspace=0.1, hspace=0.0)
        if self.show_residual:
            axs = axs_[0,:] 
        else:
            axs = axs_

        if self.show_residual:
            residual_axs = axs_[1,:]

        try:
            isinstance(axs[0], Axes)
        except:
            axs = [axs]
            residual_axs = [residual_axs]

        if n_ranges == 1:
            if self.wl_range is None:
                dwl = data_wl[-1]-data_wl[0]
                wl_low = data_wl[0] - dwl*0.025
                wl_up = data_wl[-1] + dwl*0.025
            else:
                wl_low, wl_up = self.wl_range[0], self.wl_range[1]

            axs[0].set_xlim([wl_low, wl_up])
            if self.show_residual:
                residual_axs[0].set_xlim([wl_low, wl_up])
        else:
            # how big to make the diagonal lines in axes coordinates
            # converting "points" to axes coordinates: 
            # https://stackoverflow.com/a/33638091
            t = axs[0].transAxes.transform([(0,0), (1,1)])
            t = axs[0].get_figure().get_dpi() / (t[1,1] - t[0,1]) / 72
            d = 0.5*(rcParams['xtick.major.size']*t)

            wl_ranges = [(self.wl_range[2*i], self.wl_range[2*i+1]) for i in range(n_ranges)]

            wl_l = wl_ranges[0]
            for i, (ax_l, ax_r) in enumerate(pairwise(axs)):
                kwargs = dict(transform=ax_l.transAxes, color='k', clip_on=False)
                ax_l.spines['right'].set_visible(False)
                if not ax_l.spines['left'].get_visible():
                    ax_l.yaxis.set_ticks_position('none')
                else:
                    ax_l.yaxis.tick_left()
                ax_l.set_xlim(wl_l)
                ax_l.plot((1-d, 1+d), (-d, +d), **kwargs)        
                ax_l.plot((1-d, 1+d), (1-d, 1+d), **kwargs)        

                kwargs = dict(transform=ax_r.transAxes, color='k', clip_on=False)
                ax_r.spines['left'].set_visible(False)
                ax_r.yaxis.tick_right()
                ax_r.tick_params(labelright='off') 
                wl_r = wl_ranges[1+i]
                ax_r.set_xlim(wl_r)
                ax_r.plot((-d, +d), (-d, +d), **kwargs)        
                ax_r.plot((-d, +d), (1-d, 1+d), **kwargs)        

                wl_l = wl_r

            if self.show_residual:
                t = residual_axs[0].transAxes.transform([(0,0), (1,1)])
                t = residual_axs[0].get_figure().get_dpi() / (t[1,1] - t[0,1]) / 72
                d = 0.5*(rcParams['xtick.major.size']*t)

                wl_l = wl_ranges[0]
                for i, (ax_l, ax_r) in enumerate(pairwise(residual_axs)):
                    kwargs = dict(transform=ax_l.transAxes, color='k', clip_on=False)
                    ax_l.spines['right'].set_visible(False)
                    if not ax_l.spines['left'].get_visible():
                        ax_l.yaxis.set_ticks_position('none')
                    else:
                        ax_l.yaxis.tick_left()
                    ax_l.set_xlim(wl_l)
                    ax_l.plot((1-d, 1+d), (-d, +d), **kwargs)        
                    #ax_l.plot((1-d, 1+d), (1-d, 1+d), **kwargs)        

                    kwargs = dict(transform=ax_r.transAxes, color='k', clip_on=False)
                    ax_r.spines['left'].set_visible(False)
                    ax_r.yaxis.tick_right()
                    ax_r.tick_params(labelright='off') 
                    wl_r = wl_ranges[1+i]
                    ax_r.set_xlim(wl_r)
                    ax_r.plot((-d, +d), (-d, +d), **kwargs)        
                    #ax_r.plot((-d, +d), (1-d, 1+d), **kwargs)        
                    #ax_l.spines['top'].set_visible(False)
                    ax_l.spines['top'].set_color('none')
                    ax_l.xaxis.set_ticks_position('bottom')

                    ax_r.spines['top'].set_color('none')
                    ax_r.xaxis.set_ticks_position('bottom')
                    #ax_r.spines['top'].set_visible(False)

                    ax_r.patch.set_facecolor('None')
                    ax_l.patch.set_facecolor('None')

                    wl_l = wl_r

        which = 'both'
        for ax in axs:
            if self.log_flux:
                ax.set_yscale('log')
                which = 'x'

            # Set better location of tick marks
            if self.wl_range is not None:
                set_plot_ticks(ax, which=which, prune_x='both', n_x=3)
                for tick in ax.get_xticklabels():
                    tick.set_rotation(45)
                    if self.show_residual:
                        tick.set_ha('left')
                    else:
                        tick.set_ha('right')
                    tick.set_rotation_mode("anchor")
            else:
                set_plot_ticks(ax, which=which)

        if self.show_residual:
            for ax in axs:
                ax.tick_params(labelbottom='off')

            for ax in residual_axs:
                if self.wl_range is not None:
                    set_plot_ticks(ax, prune_x='both', n_x=3)
                    for tick in ax.get_xticklabels():
                        tick.set_rotation(45)

        # Define plotting styles
        xlabel = "$\\uplambda / " + xlabel + "$"
        if not self.wl_rest:
            xlabel = xlabel + " (observed frame)"

        if self.wl_range is not None:
            fig.text(0.5, -0.04, xlabel, ha='center')
        else:
            fig.text(0.5, 0.02, xlabel, ha='center')


        ylabel = "$F_{\\uplambda} / (\\textnormal{erg} \, \
                \\textnormal{s}^{-1} \, \\textnormal{cm}^{-2} \, \
                \\textnormal{\AA}^{-1})$"
        axs[0].set_ylabel(ylabel)

        if self.show_residual:
            ylabel = "$(F_{\\uplambda}-F_{\\uplambda}^\\textnormal{mod}) / F_{\\uplambda}$"
            residual_axs[0].set_ylabel(ylabel)

        # Title of the plot is the object ID
        if self.print_ID: 
            #fig.text(0.5, 0.95, str(ID).split('_')[0].strip(), ha='center', va='top')
            plt.suptitle(str(ID).split('_')[0].strip())

        alpha_line = 0.7
        alpha_fill = 0.3

        for ax in axs:

            colors = ["red", "salmon"]
            for slic, col in zip(slices, colors):
                for s in slic:
                    if (self.draw_steps):
                        ax.step(data_wl[s],
                                data_flux[s],
                                where="mid",
                                color=col,
                                linewidth=1.50,
                                alpha=alpha_line
                                )
                    else:
                        ax.plot(data_wl[s],
                                data_flux[s],
                                color=col,
                                linewidth=2.00,
                                alpha=alpha_line
                                )

                    if (self.draw_steps):
                        FillBetweenStep.fill_between_steps(ax,
                                data_wl[s],
                                data_flux[s]-data_flux_err[s],
                                data_flux[s]+data_flux_err[s],
                                step_where="mid",
                                color=col, 
                                linewidth=0,
                                interpolate=True,
                                alpha=alpha_fill
                                )
                    else:
                        ax.fill_between(data_wl[s],
                                data_flux[s]-data_flux_err[s],
                                data_flux[s]+data_flux_err[s],
                                facecolor=col, 
                                linewidth=0,
                                interpolate=True,
                                alpha=alpha_fill
                                )


            colors = ["blue", "deepskyblue"]
            for slic, col in zip(slices_model, colors):
                for s in slic:
                    if (self.draw_steps):
                        ax.step(model_wl[s],
                                median_flux[s],
                                where="mid",
                                color=col,
                                linewidth = 1.0,
                                alpha=alpha_line
                                )
                    else:
                        ax.plot(model_wl[s],
                                median_flux[s],
                                color=col,
                                linewidth = 1.5,
                                alpha=alpha_line
                                )

                    if (self.draw_steps):
                        FillBetweenStep.fill_between_steps(ax,
                                model_wl[s],
                                lower_flux[s], 
                                upper_flux[s],
                                step_where="mid",
                                color=col, 
                                linewidth=0,
                                alpha=alpha_fill
                                )
                    else:
                        ax.fill_between(model_wl[s],
                                lower_flux[s],
                                upper_flux[s],
                                facecolor=col, 
                                linewidth=0,
                                interpolate=True,
                                alpha=alpha_fill
                                )

            # Extract and plot full SED
            if 'full sed wl' in hdulist and self.plot_full_SED:
                indices = np.arange(len(probability))
                wrand = WalkerRandomSampling(probability, keys=indices)
                rand_indices = wrand.random(self.n_SED_to_plot)

                if self.wl_rest:
                    wl_obs = hdulist['full sed wl'].data['wl'][0,:]

                    for i in rand_indices:
                        SED = hdulist['full sed'].data[i,:]
                else:
                    wl_obs = hdulist['full sed wl'].data['wl'][0,:] * z1

                    for i in rand_indices:
                        SED = hdulist['full sed'].data[i,:] / z1

                    ax.plot(wl_obs, 
                            flux_obs,
                            color="black",
                            ls="-",
                            lw=0.5,
                            alpha=0.5)

            kwargs = { 'alpha': 0.8 }

        ylim = autoscale.get_autoscale_y(axs[0], ylog=self.log_flux)
        for ax in axs:
            yl = autoscale.get_autoscale_y(ax, ylog=self.log_flux, top_margin=0.2, bottom_margin=0.1)
            ylim = [min(ylim[0], yl[0]), max(ylim[1], yl[1])]

        for ax in axs:
            ax.set_ylim(ylim)

        if self.show_residual:
            data_flux_ = data_flux[data_wl==model_wl]
            data_mask_ = data_mask[data_wl==model_wl]
            data_flux_err_ = data_flux_err[data_wl==model_wl]
            residual = (data_flux_-median_flux)/data_flux_
            residual_err = (1./data_flux_ - (data_flux_-median_flux)/data_flux_**2) * data_flux_err_

            ymax = np.amax(abs(residual))
            ymax += 0.2*ymax
            ymax = 1.1

            colors = ["darkgreen"]
            indices = np.arange(len(data_wl))
            if has_mask:
                colors = ["darkgreen", mask_color]

            for ax in residual_axs:
                ax.set_ylim([-ymax, ymax])
                kwargs = {'alpha':0.7}

                ax.plot(ax.get_xlim(), [0.,0.],
                        color="black",
                        lw=0.7,
                        zorder=5,
                        **kwargs)

                for i, col in enumerate(colors):

                    if i==0:
                        ind = indices[data_mask_]
                    elif i==1:
                        ind = indices[~data_mask_]

                    ax.errorbar(model_wl[ind],
                            residual[ind],
                            yerr=residual_err[ind],
                            color = col,
                            ls=" ",
                            elinewidth = 0.5,
                            marker='o',
                            ms=3,
                            **kwargs
                            )

        for ax in axs:

            x0, x1 = ax.get_xlim()

            y0, y1 = ax.get_ylim()
            if self.log_flux:
                y0, y1 = np.log10(y0), np.log10(y1)
            dy = y1-y0

            # Label emission lines
            if self.plot_line_labels:
                for key, label in six.iteritems(self.line_labels):
                    if self.wl_rest:
                        x = label["wl"]/wl_factor
                    else:
                        x = label["wl"]/wl_factor * z1

                    if x < x0 or x > x1:
                        continue

                    if self.resolution is not None:
                        if abs(x-prev_x) < (x/self.resolution):
                            continue

                    direction = "up"
                    if 'direction' in label:
                        direction = label['direction']

                    color = "black"
                    if 'color' in label:
                        color = label['color']

                    i1 = bisect_left(data_wl, x-2)
                    i2 = bisect_left(data_wl, x+2.)
                    if direction == "up":
                        mm = np.amax(data_flux[i1:i2+1])
                        y = mm+dy*0.1
                        ydots = mm+dy*0.07
                    elif direction == "down":
                        mm = np.amin(data_flux[i1:i2+1])
                        y = mm-dy*0.1
                        ydots = mm-dy*0.07

                    ha = 'center'
                    if 'ha' in label:
                        ha = label['ha']

                    ax.text(x, y, 
                            label["label"], 
                            fontsize=self.inset_fontsize, 
                            rotation=45,
                            color=color,
                            va='center',
                            ha=ha)

                    y0 = mm
                    ax.plot([x,x], [y0,ydots], 
                            ls=":",
                            color=color,
                            lw=1.0,
                            zorder=0)

            if print_text:

                # Print the evidence
                try:
                    ax.text(x, y, "$\\log(Z)=" + "{:.2f}".format(self.logEvidence) + "$", fontsize=self.inset_fontsize)
                except AttributeError:
                    print("ciao")

                # Print the average reduced chi-square
                try:
                    aver_chi_square = self.PPC.data['aver_chi_square'][self.PPC.data['ID'] == ID]
                    y = y1 - (y1-y0)*0.15
                    ax.text(x, y, "$\\langle\chi^2\\rangle=" + "{:.2f}".format(aver_chi_square) + "$", fontsize=self.inset_fontsize)
                except AttributeError:
                    print("`PosteriorPredictiveChecks` not computed/loaded, hence " \
                    "<chi^2> for the object `" + str(ID) + "` is not available")

                try:
                    aver_red_chi_square = self.PPC.data['aver_red_chi_square'][self.PPC.data['ID'] == ID]
                    n_data = self.PPC.data['n_used_bands'][self.PPC.data['ID'] == ID]
                    y = y1 - (y1-y0)*0.20
                    ax.text(x, y,
                            "$\\langle\chi^2/(\\textnormal{N}_\\textnormal{data}-1)\\rangle=" \
                            + "{:.2f}".format(aver_red_chi_square) + "\; \
                            (\\textnormal{N}_\\textnormal{data}=" + \
                            "{:d}".format(n_data) + ")" + "$", fontsize=self.inset_fontsize)
                except AttributeError:
                    print("`PosteriorPredictiveChecks` not computed/loaded, hence " \
                    "<chi^2_red> for the object `" + str(ID) + "` is not available")

            if y0 < 0.: plt.plot( [x0,x1], [0.,0.], color='gray', lw=1.0 )

        #plt.tick_params(labelcolor='none', top='off', bottom='off', left='off', right='off')
        #plt.xlabel(
        #plt.ylabel("$F_{\\lambda} / (\\textnormal{erg} \, \
        #        \\textnormal{s}^{-1} \, \\textnormal{cm}^{-2} \, \
        #        \\textnormal{\AA}^{-1})$", va='center', rotation='vertical')

        #fig.text(0.5, 0.02, , ha='center')
        #fig.text(0.04, 0.5, 
        #        \\textnormal{s}^{-1} \, \\textnormal{cm}^{-2} \, \
        #        \\textnormal{\AA}^{-1})$", va='center', rotation='vertical')


        #plt.tight_layout()

        name = prepare_plot_saving(plot_name)

        fig.savefig(name, dpi=None, facecolor='w', edgecolor='w',
                orientation='portrait', papertype='a4', format="pdf",
                transparent=False, bbox_inches="tight", pad_inches=0.1)
        plt.close(fig)

        hdulist.close()



from __future__ import absolute_import
def float_nsf(q,n):
    """
    Truncate a float to n significant figures.  May produce overflow in 
    very last decimal place when q < 1.  This can be removed by an extra 
    formatted print. 
    Arguments:
      q : a float
      n : desired number of significant figures
    Returns:
    Float with only n s.f. and trailing zeros, but with a possible small overflow.
    """

    import numpy as np

    sgn=np.sign(q)
    q=abs(q)
    n=int(np.log10(q/10.)) # Here you overwrite input n!
    if q<1. :
        val=q/(10**(n-1))
    else:
        val=q/(10**n)
    return str(sgn*int(val)*10.**n)

def to_precision(x,p):
    """
    returns a string representation of x formatted with a precision of p

    Based on the webkit javascript implementation taken from here:
    https://code.google.com/p/webkit-mirror/source/browse/JavaScriptCore/kjs/number_object.cpp
    """


    import math
    x = float(x)

    if x == 0.:
        return "0." + "0"*(p-1)

    out = []

    if x < 0:
        out.append("-")
        x = -x

    e = int(math.log10(x))
    tens = math.pow(10, e - p + 1)
    n = math.floor(x/tens)

    if n < math.pow(10, p - 1):
        e = e -1
        tens = math.pow(10, e - p+1)
        n = math.floor(x / tens)

    if abs((n + 1.) * tens - x) <= abs(n * tens -x):
        n = n + 1

    if n >= math.pow(10,p):
        n = n / 10.
        e = e + 1


    m = "%.*g" % (p, n)

    if 0 == 1 and (e < -2 or e >= p):
        out.append(m[0])
        if p > 1:
            out.append(".")
            out.extend(m[1:p])
        out.append('e')
        if e > 0:
            out.append("+")
        out.append(str(e))
    elif e == (p -1):
        out.append(m)
    elif e >= 0:
        out.append(m[:e+1])
        if e+1 < len(m):
            out.append(".")
            out.extend(m[e+1:])
    else:
        out.append("0.")
        out.extend(["0"]*-(e+1))
        out.append(m)

    return "".join(out)



# Taken from http://nbviewer.ipython.org/gist/tillahoffmann/f844bce2ec264c1c8cb5

from __future__ import absolute_import
import numpy as np
from scipy.spatial.distance import cdist

class gaussian_kde(object):
    """Representation of a kernel-density estimate using Gaussian kernels.

    Kernel density estimation is a way to estimate the probability density
    function (PDF) of a random variable in a non-parametric way.
    `gaussian_kde` works for both uni-variate and multi-variate data.   It
    includes automatic bandwidth determination.  The estimation works best for
    a unimodal distribution; bimodal or multi-modal distributions tend to be
    oversmoothed.

    Parameters
    ----------
    dataset : array_like
        Datapoints to estimate from. In case of univariate data this is a 1-D
        array, otherwise a 2-D array with shape (# of dims, # of data).
    bw_method : str, scalar or callable, optional
        The method used to calculate the estimator bandwidth.  This can be
        'scott', 'silverman', a scalar constant or a callable.  If a scalar,
        this will be used directly as `kde.factor`.  If a callable, it should
        take a `gaussian_kde` instance as only parameter and return a scalar.
        If None (default), 'scott' is used.  See Notes for more details.
    weights : array_like, shape (n, ), optional, default: None
        An array of weights, of the same shape as `x`.  Each value in `x`
        only contributes its associated weight towards the bin count
        (instead of 1).

    Attributes
    ----------
    dataset : ndarray
        The dataset with which `gaussian_kde` was initialized.
    d : int
        Number of dimensions.
    n : int
        Number of datapoints.
    neff : float
        Effective sample size using Kish's approximation.
    factor : float
        The bandwidth factor, obtained from `kde.covariance_factor`, with which
        the covariance matrix is multiplied.
    covariance : ndarray
        The covariance matrix of `dataset`, scaled by the calculated bandwidth
        (`kde.factor`).
    inv_cov : ndarray
        The inverse of `covariance`.

    Methods
    -------
    kde.evaluate(points) : ndarray
        Evaluate the estimated pdf on a provided set of points.
    kde(points) : ndarray
        Same as kde.evaluate(points)
    kde.pdf(points) : ndarray
        Alias for ``kde.evaluate(points)``.
    kde.set_bandwidth(bw_method='scott') : None
        Computes the bandwidth, i.e. the coefficient that multiplies the data
        covariance matrix to obtain the kernel covariance matrix.
        .. versionadded:: 0.11.0
    kde.covariance_factor : float
        Computes the coefficient (`kde.factor`) that multiplies the data
        covariance matrix to obtain the kernel covariance matrix.
        The default is `scotts_factor`.  A subclass can overwrite this method
        to provide a different method, or set it through a call to
        `kde.set_bandwidth`.

    Notes
    -----
    Bandwidth selection strongly influences the estimate obtained from the KDE
    (much more so than the actual shape of the kernel).  Bandwidth selection
    can be done by a "rule of thumb", by cross-validation, by "plug-in
    methods" or by other means; see [3]_, [4]_ for reviews.  `gaussian_kde`
    uses a rule of thumb, the default is Scott's Rule.

    Scott's Rule [1]_, implemented as `scotts_factor`, is::

        n**(-1./(d+4)),

    with ``n`` the number of data points and ``d`` the number of dimensions.
    Silverman's Rule [2]_, implemented as `silverman_factor`, is::

        (n * (d + 2) / 4.)**(-1. / (d + 4)).

    Good general descriptions of kernel density estimation can be found in [1]_
    and [2]_, the mathematics for this multi-dimensional implementation can be
    found in [1]_.

    References
    ----------
    .. [1] D.W. Scott, "Multivariate Density Estimation: Theory, Practice, and
           Visualization", John Wiley & Sons, New York, Chicester, 1992.
    .. [2] B.W. Silverman, "Density Estimation for Statistics and Data
           Analysis", Vol. 26, Monographs on Statistics and Applied Probability,
           Chapman and Hall, London, 1986.
    .. [3] B.A. Turlach, "Bandwidth Selection in Kernel Density Estimation: A
           Review", CORE and Institut de Statistique, Vol. 19, pp. 1-33, 1993.
    .. [4] D.M. Bashtannyk and R.J. Hyndman, "Bandwidth selection for kernel
           conditional density estimation", Computational Statistics & Data
           Analysis, Vol. 36, pp. 279-298, 2001.

    Examples
    --------
    Generate some random two-dimensional data:

    >>> from scipy import stats
    >>> def measure(n):
    >>>     "Measurement model, return two coupled measurements."
    >>>     m1 = np.random.normal(size=n)
    >>>     m2 = np.random.normal(scale=0.5, size=n)
    >>>     return m1+m2, m1-m2

    >>> m1, m2 = measure(2000)
    >>> xmin = m1.min()
    >>> xmax = m1.max()
    >>> ymin = m2.min()
    >>> ymax = m2.max()

    Perform a kernel density estimate on the data:

    >>> X, Y = np.mgrid[xmin:xmax:100j, ymin:ymax:100j]
    >>> positions = np.vstack([X.ravel(), Y.ravel()])
    >>> values = np.vstack([m1, m2])
    >>> kernel = stats.gaussian_kde(values)
    >>> Z = np.reshape(kernel(positions).T, X.shape)

    Plot the results:

    >>> import matplotlib.pyplot as plt
    >>> fig = plt.figure()
    >>> ax = fig.add_subplot(111)
    >>> ax.imshow(np.rot90(Z), cmap=plt.cm.gist_earth_r,
    ...           extent=[xmin, xmax, ymin, ymax])
    >>> ax.plot(m1, m2, 'k.', markersize=2)
    >>> ax.set_xlim([xmin, xmax])
    >>> ax.set_ylim([ymin, ymax])
    >>> plt.show()

    """
    def __init__(self, dataset, bw_method=None, weights=None):
        self.dataset = np.atleast_2d(dataset)
        if not self.dataset.size > 1:
            raise ValueError("`dataset` input should have multiple elements.")
        self.d, self.n = self.dataset.shape
            
        if weights is not None:
            self.weights = weights / np.sum(weights)
        else:
            self.weights = np.ones(self.n) / self.n
            
        # Compute the effective sample size 
        # http://surveyanalysis.org/wiki/Design_Effects_and_Effective_Sample_Size#Kish.27s_approximate_formula_for_computing_effective_sample_size
        self.neff = 1.0 / np.sum(self.weights ** 2)

        self.set_bandwidth(bw_method=bw_method)

    def evaluate(self, points):
        """Evaluate the estimated pdf on a set of points.

        Parameters
        ----------
        points : (# of dimensions, # of points)-array
            Alternatively, a (# of dimensions,) vector can be passed in and
            treated as a single point.

        Returns
        -------
        values : (# of points,)-array
            The values at each point.

        Raises
        ------
        ValueError : if the dimensionality of the input points is different than
                     the dimensionality of the KDE.

        """
        points = np.atleast_2d(points)

        d, m = points.shape
        if d != self.d:
            if d == 1 and m == self.d:
                # points was passed in as a row vector
                points = np.reshape(points, (self.d, 1))
                m = 1
            else:
                msg = "points have dimension %s, dataset has dimension %s" % (d,
                    self.d)
                raise ValueError(msg)

        # compute the normalised residuals
        chi2 = cdist(points.T, self.dataset.T, 'mahalanobis', VI=self.inv_cov) ** 2
        # compute the pdf
        result = np.sum(np.exp(-.5 * chi2) * self.weights, axis=1) / self._norm_factor

        return result

    __call__ = evaluate

    def scotts_factor(self):
        return np.power(self.neff, -1./(self.d+4))

    def silverman_factor(self):
        return np.power(self.neff*(self.d+2.0)/4.0, -1./(self.d+4))

    #  Default method to calculate bandwidth, can be overwritten by subclass
    covariance_factor = scotts_factor

    def set_bandwidth(self, bw_method=None):
        """Compute the estimator bandwidth with given method.

        The new bandwidth calculated after a call to `set_bandwidth` is used
        for subsequent evaluations of the estimated density.

        Parameters
        ----------
        bw_method : str, scalar or callable, optional
            The method used to calculate the estimator bandwidth.  This can be
            'scott', 'silverman', a scalar constant or a callable.  If a
            scalar, this will be used directly as `kde.factor`.  If a callable,
            it should take a `gaussian_kde` instance as only parameter and
            return a scalar.  If None (default), nothing happens; the current
            `kde.covariance_factor` method is kept.

        Notes
        -----
        .. versionadded:: 0.11

        Examples
        --------
        >>> x1 = np.array([-7, -5, 1, 4, 5.])
        >>> kde = stats.gaussian_kde(x1)
        >>> xs = np.linspace(-10, 10, num=50)
        >>> y1 = kde(xs)
        >>> kde.set_bandwidth(bw_method='silverman')
        >>> y2 = kde(xs)
        >>> kde.set_bandwidth(bw_method=kde.factor / 3.)
        >>> y3 = kde(xs)

        >>> fig = plt.figure()
        >>> ax = fig.add_subplot(111)
        >>> ax.plot(x1, np.ones(x1.shape) / (4. * x1.size), 'bo',
        ...         label='Data points (rescaled)')
        >>> ax.plot(xs, y1, label='Scott (default)')
        >>> ax.plot(xs, y2, label='Silverman')
        >>> ax.plot(xs, y3, label='Const (1/3 * Silverman)')
        >>> ax.legend()
        >>> plt.show()

        """
        if bw_method is None:
            pass
        elif bw_method == 'scott':
            self.covariance_factor = self.scotts_factor
        elif bw_method == 'silverman':
            self.covariance_factor = self.silverman_factor
        elif np.isscalar(bw_method) and not isinstance(bw_method, string_types):
            self._bw_method = 'use constant'
            self.covariance_factor = lambda: bw_method
        elif callable(bw_method):
            self._bw_method = bw_method
            self.covariance_factor = lambda: self._bw_method(self)
        else:
            msg = "`bw_method` should be 'scott', 'silverman', a scalar " \
                  "or a callable."
            raise ValueError(msg)

        self._compute_covariance()

    def _compute_covariance(self):
        """Computes the covariance matrix for each Gaussian kernel using
        covariance_factor().
        """
        self.factor = self.covariance_factor()
        # Cache covariance and inverse covariance of the data
        if not hasattr(self, '_data_inv_cov'):
            # Compute the mean and residuals
            _mean = np.sum(self.weights * self.dataset, axis=1)
            _residual = (self.dataset - _mean[:, None])
            # Compute the biased covariance
            self._data_covariance = np.atleast_2d(np.dot(_residual * self.weights, _residual.T))
            # Correct for bias (http://en.wikipedia.org/wiki/Weighted_arithmetic_mean#Weighted_sample_covariance)
            self._data_covariance /= (1 - np.sum(self.weights ** 2))
            self._data_inv_cov = np.linalg.inv(self._data_covariance)

        self.covariance = self._data_covariance * self.factor**2
        self.inv_cov = self._data_inv_cov / self.factor**2
        self._norm_factor = np.sqrt(np.linalg.det(2*np.pi*self.covariance)) #* self.n



from __future__ import absolute_import
from astropy.io import ascii
from astropy.io import fits

from .beagle_utils import is_FITS_file
        
class ObservedCatalogue(object):

    def load(self, file_name):

        """ 
        Load a catalogue of observed sources. It automatically
        detects, and loads, FITS or ASCII files depending on the suffix.

        Parameters
        ----------
        file_name : str
            Contains the file name of the catalogue.
        """

        if is_FITS_file(file_name):
            self.data = fits.open(file_name)[1].data
            self.columns = fits.open(file_name)[1].columns
        else:
            self.data = ascii.read(file_name, Reader=ascii.basic.CommentedHeader)





from __future__ import absolute_import
from __future__ import print_function
import fnmatch
import re
import os
import logging
import numpy as np
from bisect import bisect_left
from scipy.integrate import simps, cumtrapz
from scipy.interpolate import interp1d
from datetime import datetime

import sys

import matplotlib.ticker as plticker
import matplotlib as mpl
import six
from six.moves import range
from six.moves import input

import pyp_beagle.dependencies.WeightedKDE 

ID_COLUMN_LENGTH = 100

def is_integer(s):
    try:
        int(s)
        return True
    except ValueError:
        return False

def is_FITS_file(file_name):

    fits_ext = ('.fits', '.fits')
    arch_ext = ('', '.gz', '.z', '.zip')
    name = file_name.lower()
    for fExt in fits_ext:
        for aExt in arch_ext:
            suffix = fExt + aExt
            if name.endswith(suffix):
                return True

    return False

def trimFitsSuffix(file_name):

    fits_ext = ['.fits', '.fits']
    arch_ext = ['', '.gz', '.z', '.zip']
    name = file_name
    for fExt in fits_ext:
        for aExt in arch_ext:
            suffix = fExt + aExt
            if name.endswith(suffix):
                return name.split(suffix)[0]

    return False
            
def extract_IDs(table, key=None):

    ID_keys = ["ID", "ID_", "_ID"]

    if key is not None:
        return table[key]

    for name in table.dtype.names:
        if name == "ID" or name.startswith("ID") or name.endswith("ID"):
            return table[name]

    raise ValueError("Cannot extract a column containing the objects IDs!")

def extract_row(table, ID, key=None):

    if key is None:
        key='ID'

    if isinstance(table[key][0], six.string_types):
        row = table[table[key] == str(ID)]
    else:
        row = table[table[key] == int(ID)]

    if row is None or not row:
        raise ValueError("Cannot extract the row corresponding to the keyvalue '"+ID+"'")

    return row

def pause():

    try:
        wait = eval(input("PRESS ENTER TO CONTINUE."))
    except KeyboardInterrupt:
        raise SystemExit

def configure_matplotlib():

    # I used this useful vi replace command to create some of the commands below
    # :s/-\(\S\+\)\s\+:\s\+\(\S\+\)/mpl\.rcParams[\'\1\']\ =\ \2/g
    # which would turn, e.g.  
    # -ytick.minor.right    : True
    # into 
    # mpl.rcParams['ytick.minor.right'] = True

    mpl.rcParams['backend'] = 'MacOSX'

    mpl.rc('lines', 
            linewidth=1.5, linestyle='-', 
            color='blue', markeredgewidth=0.0, 
            markersize=6)

    mpl.rc('font', 
            family='serif', style='normal', 
            variant='normal', weight='normal',
            stretch='normal', size=16.0, 
            serif='Times')

    mpl.rc('text', 
            color='black', usetex=True)


    mpl.rcParams['text.usetex'] = True

    mpl.rcParams['axes.unicode_minus'] = False

    mpl.rcParams['text.latex.preamble'] = '\\usepackage{amssymb} \\usepackage{amsmath} \
            \\usepackage{upgreek} \\usepackage{txfonts} \\usepackage[utf8]{inputenc} \\DeclareUnicodeCharacter{2212}{$-$}'

    mpl.rcParams['text.latex.preview'] = True

    mpl.rc('axes', 
            facecolor='white', edgecolor='black',
            linewidth=1.5, grid=False, labelpad=4.0,
            labelcolor='black')

    mpl.rcParams['axes.formatter.limits'] = [-7., 7.]

    mpl.rcParams['xtick.top'] = True  ## draw ticks on the top side
    mpl.rcParams['xtick.bottom'] = True   ## draw ticks on the bottom side
    mpl.rcParams['xtick.major.size'] = 6.5    ## major tick size in points
    mpl.rcParams['xtick.minor.size'] = 4      ## minor tick size in points
    mpl.rcParams['xtick.major.width'] = 1.5    ## major tick width in points
    mpl.rcParams['xtick.minor.width'] = 1.2    ## minor tick width in points
    mpl.rcParams['xtick.direction'] = 'in'    ## direction: in, out, or inout
    mpl.rcParams['xtick.minor.visible'] = True   ## visibility of minor ticks on x-axis
    mpl.rcParams['xtick.major.top'] = True   ## draw x axis top major ticks
    mpl.rcParams['xtick.major.bottom'] = True   ## draw x axis bottom major ticks
    mpl.rcParams['xtick.minor.top'] = True   ## draw x axis top minor ticks
    mpl.rcParams['xtick.minor.bottom'] = True   ## draw x axis bottom minor ticks

    mpl.rcParams['ytick.left'] = True    ## draw ticks on the left side
    mpl.rcParams['ytick.right'] = True    ## draw ticks on the right side
    mpl.rcParams['ytick.major.size'] = 6.5    ## major tick size in points
    mpl.rcParams['ytick.minor.size'] = 4      ## minor tick size in points
    mpl.rcParams['ytick.major.width'] = 1.5    ## major tick width in points
    mpl.rcParams['ytick.minor.width'] = 1.2    ## minor tick width in points
    mpl.rcParams['ytick.direction'] = 'in'     ## direction: in, out, or inout
    mpl.rcParams['ytick.minor.visible'] = True  ## visibility of minor ticks on y-axis
    mpl.rcParams['ytick.major.left'] = True   ## draw y axis left major ticks
    mpl.rcParams['ytick.major.right'] = True   ## draw y axis right major ticks
    mpl.rcParams['ytick.minor.left'] = True   ## draw y axis left minor ticks
    mpl.rcParams['ytick.minor.right'] = True   ## draw y axis right minorticks

    mpl.rc('legend', 
            frameon=False, numpoints=1, handletextpad=0.2)

class BeagleDirectories(object):

    beagle_input_files = "BEAGLE-input-files"
    pypbeagle_data = os.path.join("pyp-beagle", "data")
    pypbeagle_plot = os.path.join("pyp-beagle", "plot")

    results_dir = ''

    suffix = 'BEAGLE'

    MN_suffix = '_BEAGLE_MNstats.dat'

    param_file = ''

    fontsize = 16
    inset_fontsize_fraction = 0.7

def get_files_list(results_dir=None, suffix=None):
    """ 
    Get all files ending with suffix.

    Parameters
    ----------
    results_dir : str, optional
        Directory containing the files to list. By default uses the
        RESULTS_DIR constant.

    suffix: str, optional
       Suffix of the files to list. Bu default ``BeagleDirectories.suffix``

    Returns
    -------

       file_list: str array
        List of all BEAGLE results files contained in the results directory.

       file_IDs: str array
        List of all IDs of corresponding to the BEAGLE results files contained
        in the results directory.
    """ 

    if results_dir is None:
        results_dir = BeagleDirectories.results_dir

    if suffix is None:
        suffix = BeagleDirectories.suffix + '.fits.gz'

    file_list = list()
    file_IDs = list()


    for file in sorted(os.listdir(results_dir)):
        if file.endswith(suffix) and os.path.getsize(os.path.join(results_dir, file)) > 0:
            file_list.append(file)
            file = file[0:file.find(suffix)-1]
            file_IDs.append(file)

    return file_list, file_IDs


def touch(file_name, times=None):
    with open(file_name, 'a'):
        os.utime(file_name, times)


def find_file(file_name, path):
    """ 
    Find a file in a directory.

    Parameters
    ----------
    file_name : str
        Name of the file to be found (can contain wildcards)

    path: str
        Path where to (recursively) search for file_name

    Returns
    -------
       : str
        Full path to the file_name
    """ 

    print("file_name: ", file_name)

    for root, dirs, files in os.walk(path):
        for file in files:
            if fnmatch.fnmatch(file, file_name):
                return os.path.join(root, file)

    return None

def prepare_data_saving(file_name, results_dir=None, overwrite=False):
    """ 
    Prepare directory to save a data file.  

    Parameters
    ----------
    file_name : str
        Name of the output file (without directory tree).

    results_dir : str, optional
        Directory containing the BEAGLE output files. By default uses the
        RESULTS_DIR constant.

    overwrite: bool, optional
        If rue overwrites the file, is already present, while False makes a copy of
        the original file to avoid overwriting

    Returns
    -------
    name : str
        Full path to the output file,
    """ 

    if results_dir is None:
        results_dir = BeagleDirectories.results_dir

    directory = os.path.join(results_dir, BeagleDirectories.pypbeagle_data)
    if not os.path.exists(directory):
        logging.info("Creating the directory: " + directory)
        os.makedirs(directory)

    name = os.path.join(directory, os.path.basename(file_name))
    if os.path.isfile(name) and not overwrite:
        new_name = os.path.splitext(name)[0] + datetime.now().strftime("-%Y%m%d-%H%M%S") + os.path.splitext(name)[1]
        logging.warning("The file " + name + " already exists, and it will be renamed to " + new_name)
        os.rename(name, new_name)

    return name

def getPathForPlot(file_name, results_dir=None):
    """ 
    """ 

    if results_dir is None:
        results_dir = BeagleDirectories.results_dir

    directory = os.path.join(results_dir, BeagleDirectories.pypbeagle_plot)
    return os.path.join(directory, os.path.basename(file_name))

def plot_exists(file_name, results_dir=None):
    """ 
    Check if a plot already exists in the PyP-BEAGLE tree.

    Parameters
    ----------
    file_name : str
        Name of the output file (without directory tree).

    results_dir : str, optional
        Directory containing the BEAGLE output files. By default uses the
        RESULTS_DIR constant.

    Returns
    -------
    bool
        Whether the plot aready exists or not.
    """ 

    return os.path.isfile(getPathForPlot(file_name, results_dir))

def getPathForData(file_name, results_dir=None):
    """ 
    """ 

    if results_dir is None:
        results_dir = BeagleDirectories.results_dir

    directory = os.path.join(results_dir, BeagleDirectories.pypbeagle_data)
    return os.path.join(directory, os.path.basename(file_name))

def data_exists(file_name, results_dir=None):
    """ 
    Check if a file already exists in the PyP-BEAGLE tree.

    Parameters
    ----------
    file_name : str
        Name of the output file (without directory tree).

    results_dir : str, optional
        Directory containing the BEAGLE output files. By default uses the
        RESULTS_DIR constant.

    Returns
    -------
    bool
        Whether the plot aready exists or not.
    """ 

    return os.path.isfile(getPathForData(file_name, results_dir))

def prepare_plot_saving(file_name, results_dir=None, overwrite=False):
    """ 
    Prepare directory to save a plot file.  

    Parameters
    ----------
    file_name : str
        Name of the output file (without directory tree).

    results_dir : str, optional
        Directory containing the BEAGLE output files. By default uses the
        RESULTS_DIR constant.

    overwrite: bool, optional
        If rue overwrites the file, is already present, while False makes a copy of
        the original file to avoid overwriting

    Returns
    -------
    name : str
        Full path to the output file,
    """ 

    if results_dir is None:
        results_dir = BeagleDirectories.results_dir

    directory = os.path.join(results_dir, BeagleDirectories.pypbeagle_plot)
    if not os.path.exists(directory):
        logging.info("Creating the directory: " + directory)
        os.makedirs(directory)

    name = os.path.join(directory, os.path.basename(file_name))
    if os.path.isfile(name) and not overwrite:
        new_name = os.path.splitext(name)[0] + datetime.now().strftime("-%Y%m%d-%H%M%S") + os.path.splitext(name)[1]
        logging.warning("The file " + name + " already exists, and it will be renamed to " + new_name)
        os.rename(name, new_name)

    return name

def set_font_size(ax, fontsize):

    for item in ([ax.title, ax.xaxis.label, ax.yaxis.label] +
                 ax.get_xticklabels() + ax.get_yticklabels()):
        item.set_fontsize(fontsize)


def set_plot_ticks(ax, which='both', n_x=4, n_y=4, prune_x=None, prune_y=None):
    """ 
    Set the major and minor ticks of plots.

    Parameters
    ---------
    ax : Axes instance
        The axis on which to apply yhe major/minor tick marks setting.

    n_x : int, optional
        Maximum number of intervals on the x-axis.

    n_y : int, optional
        Maximum number of intervals on the y-axis.

    prune_x : str, optional 
        Either 'lower' | 'upper' | 'both' | None, to remove
        edge ticks on the x-axis.

    prune_y : str, optional 
        Either 'lower' | 'upper' | 'both' | None, to remove
        edge ticks on the y-axis.
    """ 

    if which.lower() == 'both' or which.lower() == 'x':
        ax.xaxis.major_locations = plticker.MaxNLocator(nbins=n_x, prune=prune_x) 
        ax.xaxis.set_major_locator(ax.xaxis.major_locations)

        ax.xaxis.minor_locations = plticker.AutoMinorLocator(2)
        ax.xaxis.set_minor_locator(ax.xaxis.minor_locations)

    if which.lower() == 'both' or which.lower() == 'y':
        ax.yaxis.major_locations = plticker.MaxNLocator(nbins=n_y, prune=prune_y) 
        ax.yaxis.set_major_locator(ax.yaxis.major_locations)

        ax.yaxis.minor_locations = plticker.AutoMinorLocator(2)
        ax.yaxis.set_minor_locator(ax.yaxis.minor_locations)

def match_ID(ID_list_1, ID_list_2, sorted=False, ignore_string=None):
    """ 
    Match the ID in two catalogues.

    Parameters
    ----------
    ID_list_1 : numpy array int
        The ID for the first catalogue.

    ID_list_2 : numpy array int
        The ID for the second catalogue.

    sorted : bool, or list of bool, optional
        If True then the ID arrays are assumed to be already sorted, otherwise
        they are sorted in the routine 

    Returns
    -------
    indices_1 : numpy array int
        Array of indices such as `ID_list_1`[indices_1] = `ID_list_2`[indices_2]

    indices_2 : numpy array int
        Array of indices such as `ID_list_1`[indices_1] = `ID_list_2`[indices_2]

    """

    # Firstly, check weather ID_list_1 is longer than ID_list_2 or viceversa
    n1 = len(ID_list_1)
    n2 = len(ID_list_2)

    if n1 >= n2:
        ID_long = np.array(ID_list_1)
        n_long = n1
        ID_short = np.array(ID_list_2)
        n_short = n2
    else:
        ID_long = np.array(ID_list_2)
        n_long = n2
        ID_short = np.array(ID_list_1)
        n_short = n1

    is_int = False
    sort_long = list(range(n_long))
    sort_short = list(range(n_short))
    if issubclass(ID_long.dtype.type, np.integer) and issubclass(ID_short.dtype.type, np.integer):
        is_int = True
        if not sorted:
            # Sort the two arrays and get the indices
            sort_long = np.argsort(ID_long)
            sort_short = np.argsort(ID_short)
    else:
        ID_long = np.array(ID_long, dtype=str)
        ID_short = np.array(ID_short, dtype=str)

    match_indx_long = np.full(n_long, -1, dtype=np.int)
    match_indx_short = np.full(n_short, -1, dtype=np.int)
    mask_long = np.zeros(n_long, dtype=bool)

    for i in range(n_short):

        if is_int:
            i1 = bisect_left(ID_long[sort_long], ID_short[sort_short[i]])

            if i1 < n_long:
                if ID_long[sort_long[i1]] == ID_short[sort_short[i]]:
                    match_indx_long[i1] = sort_long[i1]
                    match_indx_short[i] = sort_short[i]
        else:
            for j in range(n_long):
                if mask_long[j]:
                    continue

                if ignore_string is not None:
                    if isinstance(ignore_string, re.compile('').__class__):
                        ID_l = ignore_string.sub('', ID_long[j])
                        ID_s = ignore_string.sub('', ID_short[i])
                    else:
                        ID_l = ID_long[j].replace(ignore_string, '')
                        ID_s = ID_short[i].replace(ignore_string, '')
                else:
                    ID_l = ID_long[j]
                    ID_s = ID_short[i]

                if ID_l == ID_s:
                    match_indx_long[i] = j
                    match_indx_short[i] = i
                    mask_long[j] = True

    if n1 >= n2:
        indices_1 = match_indx_long[match_indx_long >= 0]
        indices_2 = match_indx_short[match_indx_short >= 0]
    else:
        indices_1 = match_indx_short[match_indx_short >= 0]
        indices_2 = match_indx_long[match_indx_long >= 0]

    return indices_1, indices_2


def weighted_avg_and_std(values, weights):
    """
    Return the weighted average and standard deviation.

    Parameters
    ----------
    values : Numpy ndarray
        Contains the value of the parameter.

    weights : Numpy ndarray
        Contains the weights. Must have same shape as values.

    Returns
    -------
    average : float 
        The weighted average of the input values.

    stddev : float
        The weighted standard deviation of the input values.
    """

    average = np.average(values, weights=weights)
    variance = np.average((values-average)**2, weights=weights)  # Fast and numerically precise

    return (average, np.sqrt(variance))


def prepare_violin_plot(data, 
        weights=None, 
        min_x=None,
        max_x=None,
        nXgrid=100,
        max_interval=99.7):

    if min_x is None:
        min_x = np.min(data) 

    if max_x is None:
        max_x = np.max(data) 

    # Compute the marginal PDF through a weighted KDE
    if weights is not None:
        pdf = WeightedKDE.gaussian_kde(data, weights=weights)
    else:
        pdf = WeightedKDE.gaussian_kde(data)

    # Build a grid of value over which computing the actual PDF from its KDE
    x_grid = np.linspace(min_x, max_x, nXgrid)

    # Compute the PDF
    pdf_grid = np.array(pdf(x_grid))

    # ******************************************************************
    # NB: in this case it is correct to integrate the pdf, and not just to sum,
    # since the result of the weighted KDE algorithm is a *density*. You can
    # check this by checking the value of `pdf_norm` below, which is indeed ~1.
    # This also justify the use of `cumtrapz` below (and not just of `cumsum`,
    # as in `beagle_summary_catalogue.get1DInterval`).
    # ******************************************************************

    # Compute the PDF normalization, so its integral will be exactly 1
    pdf_norm = simps(pdf_grid, x_grid)
    pdf_grid /= pdf_norm

    # Now compute the cumulative PDF
    cumul_pdf = cumtrapz(pdf_grid, x_grid, initial=0.)
    cumul_pdf /= cumul_pdf[-1]

    # Get the interpolant of the cumulative PDF
    f_interp = interp1d(cumul_pdf, x_grid)

    # Compute the limits over which you will plot the "violin", for
    # instance showing only the cumulative PDF up to +/- 3 sigma
    intv = 0.5*(100.-max_interval)/100.
    lims = f_interp([intv,1.-intv])
    prob_lims = pdf(lims) / pdf_norm

    # The median corresponds to a cumulative probability = 0.5
    median = f_interp(0.5)

    i1 = bisect_left(x_grid, lims[0])
    i2 = bisect_left(x_grid, lims[1])

    x_violin = np.concatenate(([lims[0]], x_grid[i1+1:i2], [lims[1]]))

    y_violin = np.concatenate(([prob_lims[0]], pdf_grid[i1+1:i2], [prob_lims[1]]))


    return pdf, pdf_norm, median, x_violin, y_violin



from .beagle_utils import *
from .beagle_parsers import standard_parser
from .beagle_filters import PhotometricFilters
from .beagle_photometry import Photometry
from .beagle_pdf import PDF
from .beagle_spectra import Spectrum
from .beagle_summary_catalogue import BeagleSummaryCatalogue
from .beagle_mock_catalogue import BeagleMockCatalogue
from .beagle_residual_photometry import ResidualPhotometry
from .beagle_multinest_catalogue import MultiNestCatalogue
from .beagle_posterior_predictive_checks import PosteriorPredictiveChecks
from .beagle_spectral_indices import SpectralIndices



from __future__ import absolute_import
import os
import logging
import six.moves.configparser
from collections import OrderedDict
from scipy.interpolate import interp1d
from bisect import bisect_left
import numpy as np
from itertools import tee

import json
import matplotlib.pyplot as plt
import matplotlib.ticker as plticker
import matplotlib.gridspec as gridspec
from matplotlib import rcParams
from matplotlib.axes import Axes
#import pandas as pd
# SEABORN creates by default plots with a filled background!!
#import seaborn as sns
from astropy.io import ascii
from astropy.io import fits

import sys

import pyp_beagle.dependencies.WeightedKDE 
from pyp_beagle.dependencies.walker_random_sampling import WalkerRandomSampling
import pyp_beagle.dependencies.autoscale
from pyp_beagle.dependencies import FillBetweenStep
import pyp_beagle.dependencies.set_shared_labels  as shLab

from .significant_digits import to_precision

from .beagle_utils import BeagleDirectories, prepare_plot_saving, set_plot_ticks, plot_exists, \
        prepare_violin_plot, extract_row

from .beagle_observed_catalogue import ObservedCatalogue

# See here
# http://peak.telecommunity.com/DevCenter/PythonEggs#accessing-package-resources
# for an explanation on this approach to include data files
from pkg_resources import resource_stream
import six
from six.moves import zip
from six.moves import range

TOKEN_SEP = ":"
microJy = np.float32(1.E-23 * 1.E-06)
nanoJy = np.float32(1.E-23 * 1.E-09)

p_value_lim = 0.05

def pairwise(iterable):
    "s -> (s0,s1), (s1,s2), (s2, s3), ..."
    a, b = tee(iterable)
    next(b, None)
    return zip(a, b)


_TOKENS = ['lum', 'lumErr', 'ew', 'ewErr']
_COL_NAME = 'colName'
_LABEL = 'label'
_SEP = ':'

class SpectralIndicesCatalogue(ObservedCatalogue):

    def configure(self, file_name):

        self.line_config = OrderedDict()

        for line in open(file_name):
            if line.strip() and not line.startswith("#"):
                _line = line

                _key = _LABEL + _SEP
                if _key in line:
                    _label = line.split(_key)[1].split()[0]
                    self.line_config[_label] = OrderedDict()

                    for t in _TOKENS:
                        _key = t + _SEP + _COL_NAME + _SEP
                        if _key in line:
                            _value = line.split(_key)[1].split()[0]
                            self.line_config[_label][t] = _value



class SpectralIndices(object):

    def __init__(self, **kwargs):

        with open(kwargs.get('line_labels_json')) as f:    
            self.line_list = json.load(f, object_pairs_hook=OrderedDict)

        self.inset_fontsize = BeagleDirectories.inset_fontsize_fraction * BeagleDirectories.fontsize

        self.observed_catalogue = SpectralIndicesCatalogue()

        self.key = kwargs.get('ID_key') 

        self.plot_log_flux = kwargs.get('plot_log_flux')

        self.print_values = kwargs.get('print_line_values')

        self.max_interval = kwargs.get('max_interval')


    def plot_line_fluxes(self, 
            ID, replot=False, 
            title=False, letter=None, signif_digits=1):

        suffix = ""
        #if self.print_values:
        #    suffix = "_printed_values"

        plot_name = str(ID) + '_BEAGLE_spectral_indices' + suffix + ".pdf"

        # Check if the plot already exists
        if plot_exists(plot_name) and not replot:
            logging.warning('The plot "' + plot_name + '" already exists. \n Exiting the function.')
            return

        # From the (previously loaded) observed catalogue select the row
        # corresponding to the input ID
        observation = extract_row(self.observed_catalogue.data, ID, key=self.key)

        fig = plt.figure(figsize=(12, 3))
        ax = fig.add_subplot(1, 1, 1)

        # Open the file containing BEAGLE results
        fits_file = os.path.join(BeagleDirectories.results_dir,
                str(ID) + '_' + BeagleDirectories.suffix + '.fits.gz')

        hdulist = fits.open(fits_file)

        # Read model fluxes
        model_fluxes = hdulist['spectral indices'].data
        n_lines = len(self.line_list)

        # Read the posterior probability
        probability = hdulist['posterior pdf'].data['probability']

        width = 0.5

        minY_values = np.zeros(n_lines) ; maxY_values = np.zeros(n_lines)
        _observed_fluxes = np.zeros(n_lines) ; _observed_flux_errors = np.zeros(n_lines)
        _model_fluxes = np.zeros(n_lines)
        for i, (key, value) in enumerate(six.iteritems(self.line_list)):

            X = i+1
            _line_conf = self.observed_catalogue.line_config[key]
            if 'lum' in _line_conf:
                _col_name = _line_conf['lum']
                _err_col_name = _line_conf['lumErr']
            elif 'ew' in _line_conf:
                _col_name = _line_conf['ew']
                _err_col_name = _line_conf['ewErr']

            _observed_flux = observation[_col_name]
            _observed_fluxes[i] = _observed_flux 

            _observed_flux_err = observation[_err_col_name]
            _observed_flux_errors[i] = _observed_flux_err

            _label = value["label"]

            # This function provides you with all the necessary info to draw violin plots
            _model_flux = model_fluxes[key]
            kde_pdf, pdf_norm, median_flux, x_plot, y_plot = prepare_violin_plot(_model_flux, weights=probability) 
            _model_fluxes[i] = median_flux

            _max_y = np.max(y_plot)
            w = width / _max_y

            y_grid = np.full(len(x_plot), X)
            _lim_y = kde_pdf(median_flux)/pdf_norm * w

            kwargs = {'alpha':0.8}
            ax.errorbar(X,
                    _observed_flux, 
                    yerr=_observed_flux_err,
                    color="dodgerblue",
                    ls=" ",
                    marker=" ",
                    zorder=5,
                    **kwargs)

            ax.plot(X,
                    _observed_flux, 
                    color="dodgerblue",
                    ls=" ",
                    marker="D",
                    markeredgewidth=0.,
                    markersize = 8,
                    zorder=3,
                    **kwargs)

            kwargs = {'color':'tomato', 'alpha':0.7, 'edgecolor':'black', 'linewidth':0.2}
            ax.fill_betweenx(x_plot,
                    y_grid - y_plot*w,
                    y_grid + y_plot*w,
                    zorder=2,
                    **kwargs
                    )

            ax.plot([X-_lim_y, X+_lim_y],
                    [median_flux, median_flux],
                    color = 'black',
                    zorder=2,
                    linewidth = 0.2
                    )

            ax.plot(X,
                    median_flux,
                    color = 'black',
                    marker = "o",
                    markersize = 5,
                    zorder=4,
                    alpha = 0.7
                    )

            _all = np.concatenate((_observed_flux-_observed_flux_err, x_plot))
            _min = np.amin(_all[_all > 0.])
            minY_values[i] = _min

            _all = np.concatenate((_observed_flux+_observed_flux_err, x_plot))
            _max = np.amax(_all[_all > 0.])
            maxY_values[i] = _max

        minY = np.amin(minY_values)
        maxY = np.max(maxY_values)
        if self.print_values:
            _factor = 0.15
        else:
            _factor = 0.10

        if self.plot_log_flux:
            dY = np.log10(maxY)-np.log10(minY)
            minY = 10.**(np.log10(minY)-_factor*dY)
            maxY = 10.**(np.log10(maxY)+_factor*dY)
        else:
            dY = maxY-minY
            minY -= _factor*dY
            maxY += _factor*dY
            
        ax.set_ylim(minY, maxY) 
        
        # Set better location of tick marks
        set_plot_ticks(ax, n_x=5)


        if self.plot_log_flux:
            ax.set_ylabel("$\\log(\\textnormal{F}/\\textnormal{erg} \; \
                \\textnormal{s}^{-1} \, \\textnormal{cm}^{-2})$")
        else:
            ax.set_ylabel("$\\textnormal{F}/\\textnormal{erg} \; \
                \\textnormal{s}^{-1} \, \\textnormal{cm}^{-2}$")


        plt.tick_params(
            axis='x',          # changes apply to the x-axis
            which='both',      # both major and minor ticks are affected
            top='off',      # ticks along the bottom edge are off
            bottom='off')         # ticks along the top edge are off

        xticks = list(range(1, len(self.line_list)+1))
        ax.set_xticks(xticks)

        ticklabels = list()
        alpha = 0.6
        minY, maxY = ax.get_ylim() ; dY = maxY - minY
        _factor = 0.03  ; _init_fact = 0.04
        j = 1
        _fontsize = min(self.inset_fontsize, self.inset_fontsize/(0.075*n_lines))
        for i, (line_key, line_value) in enumerate(six.iteritems(self.line_list)):
            X = i+1
            ticklabels.append(line_value["label"])
            _variab_fact = _init_fact * j
            if self.plot_log_flux:
                dY = np.log10(maxY) - np.log10(minY)
                Y = 10.**(np.log10(minY_values[i]) - _factor*dY)
                Y_t0 = 10.**(np.log10(maxY) - _variab_fact*dY)
                Y_t1 = 10.**(np.log10(maxY) - _variab_fact*dY - 1.4*_init_fact*dY)
            else:
                Y = minY_values[i] - _factor*dY
                Y_t0 = maxY - _variab_fact*dY
                Y_t1 = maxY - _variab_fact*dY - 1.4*_init_fact*dY

            ax.plot([X,X], [minY, Y],
                    color="black",
                    ls=":",
                    zorder=2,
                    alpha=alpha)

            if self.print_values:
                _val = _observed_fluxes[i]
                if not _val > 0.:
                    continue
                _val_err = _observed_flux_errors[i]
                _n = int(np.floor(np.log10(_val)))
                _norm = 10.**_n
                #_text = '$' + to_precision(_val/_norm, signif_digits+1) + '\\pm' + to_precision(_val_err/_norm, signif_digits) + '\\; 10^{' + str(_n) + '}$'
                _text = '$' + to_precision(_val/_norm, signif_digits+2) + '\\pm' + to_precision(_val_err/_norm, signif_digits) + '$'
                ax.text(X, Y_t0,
                        _text,
                        horizontalalignment='center',
                        verticalalignment='center',
                        color="black", 
                        fontsize=_fontsize)

                _val = _model_fluxes[i]
                _text = '$' + to_precision(_val/_norm, signif_digits+2) + '$'
                ax.text(X, Y_t1,
                        _text,
                        horizontalalignment='center',
                        verticalalignment='center',
                        color="black", 
                        fontsize=_fontsize)

            j += 1
            if j%4 == 0:
                j = 1

        ax.set_xticklabels(ticklabels, rotation=45, rotation_mode="anchor", ha='right')
        #zed = [tick.label.set_fontsize(fontsize*factor) for tick in ax.yaxis.get_major_ticks()]

        if self.plot_log_flux:
            ax.set_yscale('log')

        if title:
            plt.title(title)

        if letter is not None:
            ax.text(0.010, 0.940, '('+letter+')',
                    horizontalalignment='left',
                    verticalalignment='center',
                    color="black", 
                    transform=ax.transAxes)

        name = prepare_plot_saving(plot_name)

        fig.savefig(name, dpi=None, facecolor='w', edgecolor='w',
                orientation='portrait', papertype='a4', format="pdf",
                transparent=False, bbox_inches="tight", pad_inches=0.1)

        plt.close(fig)
        hdulist.close()





from __future__ import absolute_import
import six.moves.copyreg
import types

def _pickle_method(method):
	func_name = method.__func__.__name__
	obj = method.__self__
	cls = method.__self__.__class__
	if func_name.startswith('__') and not func_name.endswith('__'): #deal with mangled names
		cls_name = cls.__name__.lstrip('_')
		func_name = '_' + cls_name + func_name
	return _unpickle_method, (func_name, obj, cls)

def _unpickle_method(func_name, obj, cls):
	for cls in cls.__mro__:
		try:
			func = cls.__dict__[func_name]
		except KeyError:
			pass
		else:
			break
	return func.__get__(obj, cls)

six.moves.copyreg.pickle(types.MethodType, _pickle_method, _unpickle_method)




from __future__ import absolute_import
from __future__ import print_function
import logging
import os
from scipy.interpolate import interp1d
from collections import OrderedDict
from bisect import bisect_left
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.ticker as plticker
#import pandas as pd
# SEABORN creates by default plots with a filled background!!
#import seaborn as sns
from astropy.io import ascii
from astropy.io import fits

import sys
import pyp_beagle.dependencies.WeightedKDE 
from pyp_beagle.dependencies.walker_random_sampling import WalkerRandomSampling

from .beagle_utils import BeagleDirectories, prepare_plot_saving, set_plot_ticks, \
        prepare_violin_plot, plot_exists, pause, extract_row, is_FITS_file
from .beagle_filters import PhotometricFilters
from .beagle_summary_catalogue import BeagleSummaryCatalogue
from .beagle_residual_photometry import ResidualPhotometry
from .beagle_multinest_catalogue import MultiNestCatalogue
from .beagle_posterior_predictive_checks import PosteriorPredictiveChecks
from .beagle_observed_catalogue import ObservedCatalogue
from six.moves import range


Jy = np.float32(1.E-23)
microJy = np.float32(1.E-23 * 1.E-06)
nanoJy = np.float32(1.E-23 * 1.E-09)
c_light = 2.99792e+18 # Ang/s

p_value_lim = 0.05



class PhotometricCatalogue(ObservedCatalogue):

    def extract_fluxes(self, filters, ID, key='ID', aper_corr=1.):
        """ 
        Extract fluxes and error fluxes for a single object (units are Jy).

        Parameters
        ----------
        filters : class
            Contains the photometric filters

        ID : int, str
            Contains the object ID

        Returns    
        -------
        flux : array
            In units of Jy

        flux_error : array 
            In units of Jy

        Notes
        -----
        The routine also adds in quadrature the minimum relative error defined int he filters class.

        """

        flux = np.zeros(filters.n_bands, np.float32)
        flux_err = np.zeros(filters.n_bands, np.float32)

        row = extract_row(self.data, ID, key=key)

        for j in range(filters.n_bands):

            # observed flux and its error
            name = filters.data['flux_colName'][j]
            if not name:
                flux[j] = -99.
                flux_err[j] = -99.
                continue

            flux[j] = row[name] * aper_corr * filters.units / Jy

            name = filters.data['flux_errcolName'][j]
            flux_err[j] = row[name] * aper_corr * filters.units / Jy

            if flux_err[j] > 0.:
                # if defined, add the minimum error in quadrature
                flux_err[j] = (np.sqrt((flux_err[j]/flux[j])**2 +
                    np.float32(filters.data['min_rel_err'][j])**2) *
                    abs(flux[j]))

        return flux, flux_err

class Photometry:

    def __init__(self, **kwargs):
        
        self.inset_fontsize = BeagleDirectories.inset_fontsize_fraction * BeagleDirectories.fontsize

        self.filters = PhotometricFilters()

        self.observed_catalogue = PhotometricCatalogue()

        self.summary_catalogue = BeagleSummaryCatalogue()

        self.multinest_catalogue = MultiNestCatalogue()

        self.residual = ResidualPhotometry()

        self.PPC = PosteriorPredictiveChecks()

        self.key = kwargs.get('ID_key', 'ID')
        
        self.x_log = kwargs.get('plot_log_wl', False)

        self.plot_full_SED = kwargs.get('plot_full_SED', False)

        self.plot_MAP_SED = kwargs.get('plot_MAP_SED', False)

        self.log_flux = kwargs.get('log_flux', False)

        self.plot_filter_labels = kwargs.get('plot_filter_labels', False)

        self.flux_units = kwargs.get('flux_units', 'microJy')

        self.single_solutions = None
        if kwargs.get('plot_single_solution') is not None:
            self.single_solutions = OrderedDict()
            with fits.open(kwargs.get('plot_single_solution')) as f:
                self.single_solutions['ID'] = f[1].data['ID']
                self.single_solutions['row'] = f[1].data['row_index']

    def plot_marginal(self, ID, max_interval=99.7, 
            print_text=False, print_title=False, replot=False, show=False, units='nanoJy',
            SED_prob_log_scale=False, n_SED_to_plot=10):
        """ 
        Plot the fluxes predicted by BEAGLE.

        The fluxes here considered are those predicted by BEAGLE, given the
        posterior distribution of the model parameters. These are *not*
        replicated data.

        Parameters
        ----------
        ID : int
            ID of the galaxy whose marginal photometry will be plotted.

        max_interval : float, optional
            The marginal photometry is shown to include `max_interval`
            probability, e.g. `max_interval` = 68. will show the 68 % (i.e.
            '1-sigma') (central) credible region of the marginal photometry.

        print_text : bool, optional
            Whether to print further information on the plot, such as
            chi-square, p-value, or leave it empty and neat.

        print_text : bool, optional
            Whether to print the object ID on the top of the plot.

        replot: bool, optional
            Whether to redo the plot, even if it already exists
        """

        if self.flux_units == 'milliJy':
            flux_factor = 1.E+03
            ylabel = "$f_{\\nu}/\\textnormal{mJy}$"
        elif self.flux_units == 'microJy':
            flux_factor = 1.E+06
            ylabel = "$f_{\\nu}/\\upmu\\textnormal{Jy}$"
        elif self.flux_units == 'nanoJy':
            flux_factor = 1.E+09
            ylabel = "$f_{\\nu}/\\textnormal{nJy}$"
        else:
            raise ValueError("Flux units `" + self.flux_units + "` not recognised!")

        # Name of the output plot
        plot_name = str(ID)+'_BEAGLE_marginal_SED_phot.pdf'

        # Check if the plot already exists
        if plot_exists(plot_name) and not replot and not show:
            logging.warning('The plot "' + plot_name + '" already exists. \n Exiting the function.')
            return

        # From the (previously loaded) observed catalogue select the row
        # corresponding to the input ID
        observation = extract_row(self.observed_catalogue.data, ID, key=self.key)

        # Check if you need to apply an aperture correction to the catalogue fluxes
        if 'aper_corr' in self.observed_catalogue.data.dtype.names:
            aper_corr = 10.**(-0.4*observation[0]['aper_corr'])
        else:
            aper_corr = 1.

        # Put observed photometry and its error in arrays
        obs_flux, obs_flux_err = self.observed_catalogue.extract_fluxes(self.filters, ID, key=self.key)
        obs_flux *= flux_factor
        obs_flux_err *= flux_factor

        fig = plt.figure()
        ax = fig.add_subplot(1, 1, 1)

        # Open the file containing BEAGLE results
        fits_file = os.path.join(BeagleDirectories.results_dir,
                str(ID) + '_' + BeagleDirectories.suffix + '.fits.gz')

        hdulist = fits.open(fits_file)

        # Consider only the extension containing the predicted model fluxes
        old_API = False
        try:
            model_sed = hdulist['marginal photometry']
            old_API = True
        except:
            model_sed = hdulist['apparent magnitudes']

        probability = hdulist['posterior pdf'].data['probability']

        n_bands = len(obs_flux)
        median_flux = np.zeros(n_bands)
        pdf_norm = np.zeros(n_bands)
        _max_y = np.zeros(n_bands)
        min_flux = np.zeros(n_bands)
        max_flux = np.zeros(n_bands)

        y_plot = list(range(n_bands))
        x_plot = list(range(n_bands))

        has_pdf = list(range(n_bands))
        kde_pdf = list(range(n_bands))
        nXgrid = 1000

        if self.x_log:
            wl_eff = np.log10(self.filters.data['wl_eff'])
        else:
            wl_eff = np.array(self.filters.data['wl_eff'])

        # Sort wl_eff array
        sor = np.argsort(wl_eff)

        obs_flux, obs_flux_err, wl_eff = obs_flux[sor], obs_flux_err[sor], wl_eff[sor]

        ok = np.where(obs_flux_err > 0.)[0]

        width = 5*np.min(wl_eff[1:]-wl_eff[0:-1])

        kwargs = {'color':'tomato', 'alpha':0.8, 'edgecolor':'black', 'linewidth':0.2}

        for i in range(n_bands):

            if old_API:
                band_name = self.filters.data['label'][sor[i]]
                xdata = model_sed.data['_'+band_name+'_'] / Jy * flux_factor
            else:
                band_name = self.filters.data['name'][sor[i]] + "_APP"
                xdata = 10.**(0.4*(8.9-model_sed.data[band_name])) * flux_factor

            min_x = np.min(xdata)
            max_x = np.max(xdata)

            min_flux[i] = np.min([min_x, obs_flux[i]-obs_flux_err[i]])
            max_flux[i] = np.max([max_x, obs_flux[i]+obs_flux_err[i]])

            # if min_x == max_x, then you can not use weighted KDE, since you
            # just have one value for the x...this usually happens bacause of
            # IGM absorption, which absorbs the flux blue-ward 1216 AA, making
            # all flux = 0
            if min_x == max_x:
                has_pdf[i] = False
                median_flux[i] = min_x
                continue

            # Compute the marginal PDF through a weighted KDE
            has_pdf[i] = True

            # This function provides you with all the necessary info to draw violin plots
            kde_pdf[i], pdf_norm[i], median_flux[i], x_plot[i], y_plot[i] = prepare_violin_plot(xdata, weights=probability) 

            _max_y[i] = np.max(y_plot[i])

        delta_wl = wl_eff[1:]-wl_eff[0:-1]
        delta_wl = np.concatenate(([delta_wl[0]], delta_wl))
        delta_wl /=  2.

        for i in range(n_bands):

            dwl = delta_wl[i]
            if i > 1:
                dwl = np.min(delta_wl[i-1:i])

            if has_pdf[i]:

                w = 0.4 * dwl / _max_y[i]

                y_grid = np.full(len(x_plot[i]), wl_eff[i])
                _lim_y = kde_pdf[i](median_flux[i])/pdf_norm[i] * w

                ax.fill_betweenx(x_plot[i],
                        y_grid - y_plot[i]*w,
                        y_grid + y_plot[i]*w,
                        **kwargs
                        )

                ax.plot([wl_eff[i]-_lim_y, wl_eff[i]+_lim_y],
                        [median_flux[i], median_flux[i]],
                        color = 'black',
                        linewidth = 0.2
                        )

            ax.plot(wl_eff[i],
                    median_flux[i],
                    color = 'black',
                    marker = "o",
                    markersize = 5,
                    zorder=3,
                    alpha = 0.6
                    )


        # Plot the full SED
        if 'full sed wl' in hdulist: 

            if self.plot_full_SED or self.plot_MAP_SED:

                _n_SED_to_plot = 0
                if self.plot_full_SED: 
                    _n_SED_to_plot = n_SED_to_plot

                if self.plot_MAP_SED:
                    _n_SED_to_plot += 1

                wl = hdulist['full sed wl'].data['wl'][0,:]
                redshifts = hdulist['galaxy properties'].data['redshift']

                if SED_prob_log_scale:
                    max_prob = np.log10(np.amax(probability))
                    min_prob = np.log10(np.amin(probability))
                else:
                    max_prob = np.amax(probability)
                    min_prob = np.amin(probability)

                indices = np.arange(len(probability))

                wrand = WalkerRandomSampling(probability, keys=indices)
                rand_indices = wrand.random(n_SED_to_plot)

                for j in range(_n_SED_to_plot):

                    if self.plot_MAP_SED and j == 0:
                        i = np.argmax(probability)
                        color = "black"
                        lw=1.2
                        max_alpha = 0.8
                    else:
                        i = rand_indices[j]
                        color = "black"
                        lw=0.5
                        max_alpha = 0.4

                    SED = hdulist['full sed'].data[i,:]

                    z = 0.
                    if redshifts[i] > 0.:
                        z = redshifts[i]

                    # Redshift the SED and wl
                    flux_obs = SED / (1.+z)
                    wl_obs = wl * (1.+z)

                    # Convert F_lambda [erg s^-1 cm^-2 A^-1] ----> F_nu [erg s^-1 cm^-2 Hz^-1]
                    flux_obs = (wl_obs)**2/c_light*flux_obs

                    # Scale to nanoJy
                    flux_obs = flux_obs / Jy * flux_factor

                    prob = probability[i]
                    if SED_prob_log_scale:
                        alpha = (np.log10(prob)-min_prob)/(max_prob-min_prob)
                    else:
                        alpha = (prob-min_prob)/(max_prob-min_prob)

                    if self.x_log:
                        wl_obs = np.log10(wl_obs)

                    alpha=1.
                    ax.plot(wl_obs, 
                            flux_obs,
                            color=color,
                            ls="-",
                            lw=lw,
                            alpha=alpha*max_alpha)


        # Determine min and max values of y-axis
        yMax = np.max(max_flux)
        yMin = np.min(np.concatenate((obs_flux[ok], min_flux)))

        dY = yMax-yMin

        yMax += dY * 0.1
        if self.log_flux:
            yMin = 0.
        else:
            yMin -= dY * 0.1

        ax.set_ylim([yMin, yMax])

        x0 = wl_eff[0]
        x1 = wl_eff[-1]
        dx = x1-x0
        ax.set_xlim([x0-0.05*dx, x1+0.05*dx])

        x0, x1 = ax.get_xlim()
        if yMin < 0.: plt.plot( [x0,x1], [0.,0.], color='gray', lw=0.8 )

        # Plot labels of photometric filters
        if self.plot_filter_labels:
            for i in range(n_bands):
                x = wl_eff[i]
                y0, y1 = ax.get_ylim() ; dy = y1-y0
                y = np.min([max_flux[i]+dy*0.2, y1-0.05*dy])

                if old_API:
                    band_name = self.filters.data['label'][sor[i]]
                else:
                    band_name = self.filters.data['name'][sor[i]]

                band_name.replace("_", "") ; band_name = "$" + band_name + "$"

                rotation = 0
                if len(band_name) >= 5:
                    rotation = 45

                ax.text(x, y, 
                        band_name, 
                        fontsize=self.inset_fontsize, 
                        rotation=rotation,
                        color='black',
                        va='center',
                        ha='center')

                ydots = y-0.025*dy
                y = max_flux[i]
                ax.plot([x,x], [y,ydots], 
                        ls=":",
                        color='darkgrey',
                        lw=1.0,
                        zorder=0)


        # Define plotting styles
        if self.x_log:
            ax.set_xlabel("$\\log (\lambda_\\textnormal{eff} / \\textnormal{\AA})$")
        else:
            ax.set_xlabel("$\\lambda_\\textnormal{eff} / \\textnormal{\AA}$")

        ax.set_ylabel(ylabel)

        # Set better location of tick marks
        set_plot_ticks(ax, n_x=5)

        kwargs = {'alpha':0.7}

        plt.errorbar(wl_eff[ok], 
                obs_flux[ok], 
                yerr = obs_flux_err[ok],
                color = "dodgerblue",
                ls = " ",
                marker = "D",
                markeredgewidth = 0.,
                markersize = 8,
                elinewidth=1.0,
                capsize=3,
                **kwargs)

        if self.single_solutions is not None:
            row =  self.single_solutions['row'][self.single_solutions['ID']==ID]
            solution = np.zeros(n_bands, dtype=np.float32)
            for i, band_name in enumerate((self.filters.data['label'][sor])):
                solution[i] = model_sed.data['_'+band_name+'_'][row] / nanoJy

            ax.plot(wl_eff,
                    solution,
                    color = 'green',
                    marker = "*",
                    ls="",
                    markersize = 10,
                    alpha = 0.7
                    )

        if self.log_flux:
            ax.set_yscale('symlog')
            which = 'x'

        # Title of the plot is the object ID
        if print_title: plt.title(str(ID))

        # Location of printed text
        x0, x1 = ax.get_xlim()
        x = x0 + (x1-x0)*0.03
        y0, y1 = ax.get_ylim()
        y = y1 - (y1-y0)*0.10

        if print_text:

            # Print the evidence
            try:
                ax.text(x, y, "$\\log(Z)=" + "{:.2f}".format(self.logEvidence) + "$", fontsize=10 )
            except AttributeError:
                print("ciao")

            # Print the average reduced chi-square
            try:
                row = extract_row(self.PPC.data, ID, key=self.key)
                aver_chi_square = row['aver_chi_square']
                y = y1 - (y1-y0)*0.15
                ax.text(x, y, "$\\langle\chi^2\\rangle=" + "{:.2f}".format(aver_chi_square) + "$", fontsize=10 )
            except AttributeError:
                print("`PosteriorPredictiveChecks` not computed/loaded, hence " \
                "<chi^2> for the object `" + str(ID) + "` is not available")

            try:
                row = extract_row(self.PPC.data, ID, key=self.key)
                aver_red_chi_square = row['aver_red_chi_square']
                n_data = row['n_used_bands']
                y = y1 - (y1-y0)*0.20
                ax.text(x, y,
                        "$\\langle\chi^2/(\\textnormal{N}_\\textnormal{data}-1)\\rangle=" \
                        + "{:.2f}".format(aver_red_chi_square) + "\; \
                        (\\textnormal{N}_\\textnormal{data}=" + \
                        "{:d}".format(n_data) + ")" + "$", fontsize=10 )
            except AttributeError:
                print("`PosteriorPredictiveChecks` not computed/loaded, hence " \
                "<chi^2_red> for the object `" + str(ID) + "` is not available")

        if y0 < 0.: plt.plot( [x0,x1], [0.,0.], color='gray', lw=1.0 )

        if show:
            plt.show()
        else:
            name = prepare_plot_saving(plot_name)

            fig.savefig(name, dpi=None, facecolor='w', edgecolor='w',
                    orientation='portrait', papertype='a4', format="pdf",
                    transparent=False, bbox_inches="tight", pad_inches=0.1)

        plt.close(fig)

        hdulist.close()

    def plot_replicated_data(self, ID, max_interval=99.7, n_replic_to_plot=16,
            print_text=False, replot=False):    
        """ 
        Plot the replicated data.

        Parameters
        ----------
        ID : int
            ID of the galaxy whose marginal photometry will be plotted.

        max_interval : float, optional
            The marginal photometry is shown to include `max_interval`
            probability, e.g. `max_interval` = 68. will show the 68 % (i.e.
            '1-sigma') (central) credible region of the marginal photometry.

        n_replic_to_plot: int, optional
            The number of replicated data that will be plotted. It can be given
            as a single number, or as a pair (n_x, n_y), in which case the
            total number of replicated data plotted will be n_x * n_y

        print_text : bool, optional
            Whether to print further information on the plot, such as
            chi-square, p-value, or leave it empty and neat.

        replot: bool, optional
            Whether to redo the plot, even if it already exists
        """

        if self.flux_units == 'milliJy':
            flux_factor = 1.E+03
            ylabel = "$f_{\\nu}/\\textnormal{mJy}$"
        elif self.flux_units == 'microJy':
            flux_factor = 1.E+06
            ylabel = "$f_{\\nu}/\\upmu\\textnormal{Jy}$"
        elif self.flux_units == 'nanoJy':
            flux_factor = 1.E+09
            ylabel = "$f_{\\nu}/\\textnormal{nJy}$"
        else:
            raise ValueError("Flux units `" + self.flux_units + "` not recognised!")

        # Name of the output plot
        plot_name = str(ID)+'_BEAGLE_replic_data_phot.pdf'

        # Check if the plot already exists
        if plot_exists(plot_name) and not replot:
            logging.warning('The plot "' + plot_name + '" already exists. \n Exiting the function.')
            return

        n_replic_to_plot = np.array(n_replic_to_plot)
        if n_replic_to_plot.size == 1:
            n_plot_x = int(np.sqrt(n_replic_to_plot))
            n_plot_y = n_plot_x
        else:
            n_plot_x = n_replic_to_plot[0]
            n_plot_y = n_replic_to_plot[1]

        # From the (previously loaded) observed catalogue select the row
        # corresponding to the input ID
        observation = extract_row(self.observed_catalogue.data, ID, key=self.key)

        # Check if you need to apply an aperture correction to the catalogue fluxes
        if 'aper_corr' in self.observed_catalogue.data.dtype.names:
            aper_corr = 10.**(-0.4*observation[0]['aper_corr'])
        else:
            aper_corr = 1.

        # Put observed photometry and its error in arrays
        obs_flux, obs_flux_err = self.observed_catalogue.extract_fluxes(self.filters, ID, key=self.key)

        # Sort wl_eff array
        wl_eff = np.array(self.filters.data['wl_eff'])
        sor = np.argsort(wl_eff)

        obs_flux, obs_flux_err, wl_eff = obs_flux[sor], obs_flux_err[sor], wl_eff[sor]

        obs_flux *= flux_factor
        obs_flux_err *= flux_factor

        # Consider only those bands with measurements!
        ok = np.where(obs_flux_err > 0.)[0]

        # Open the file containing BEAGLE results
        fits_file = os.path.join(BeagleDirectories.results_dir,
                str(ID)+'_BEAGLE.fits.gz')

        model_hdu = fits.open(fits_file)
        model_sed = model_hdu['marginal photometry']

        # Open the file containing the replicated data
        fits_file = os.path.join(BeagleDirectories.results_dir,
                BeagleDirectories.pypbeagle_data,
                str(ID)+'_BEAGLE_replic_data.fits.gz')

        replic_hdu = fits.open(fits_file)
        replic_data = replic_hdu[1]

        n_replicated = replic_data.data.field(0).size

        # the first column is the ID, so the number of bands is n-1
        n_bands = len(replic_data.columns.names)-1

        indices = replic_data.data['row_index']
        noiseless_flux = np.zeros((n_bands, indices.size))

        for i, band_name in enumerate((self.filters.data['label'][sor])):
            noiseless_flux[i, :] = model_sed.data['_'+band_name+'_'] / Jy * flux_factor

        # Compute the p-value band-by-band
        p_value_bands = np.zeros(n_bands)
        replic_fluxes = np.zeros((n_bands, n_replicated))
        for i, band_name in enumerate((self.filters.data['label'][sor])):
            
            if obs_flux_err[i] > 0.:
                obs_discr = (obs_flux[i].repeat(n_replicated)-noiseless_flux[i, :])**2 / obs_flux_err[i].repeat(n_replicated)**2
                repl_discr = (replic_data.data['_'+band_name+'_']*flux_factor - noiseless_flux[i, :])**2 / obs_flux_err[i].repeat(n_replicated)**2

                replic_fluxes[i,:] = replic_data.data['_'+band_name+'_'] * flux_factor
                p_value_bands[i] = 1. * np.count_nonzero((repl_discr >
                    obs_discr)) / n_replicated

        markers = np.array("o").repeat(n_bands)
        loc = np.where(p_value_bands <= p_value_lim)[0]
        markers[loc] = "o"
        print("p_value_bands: ", p_value_bands)

        ext_obs_flux = obs_flux.reshape(n_bands, 1).repeat(n_replicated, 1)
        ext_obs_flux_err = obs_flux_err.reshape(n_bands, 1).repeat(n_replicated, 1)

        obs_discr = np.sum((ext_obs_flux[ok,:]-noiseless_flux[ok,:])**2 / ext_obs_flux_err[ok,:]**2, axis=0)
        repl_discr = np.sum((replic_fluxes[ok,:]-noiseless_flux[ok,:])**2 / ext_obs_flux_err[ok,:]**2, axis=0)

        p_value = 1. * np.count_nonzero((repl_discr >
            obs_discr)) / n_replicated
        
        print("p_value: ", p_value)

        median_flux = np.zeros(n_bands)
        pdf_norm = np.zeros(n_bands)
        _max_y = np.zeros(n_bands)
        max_abs_flux = np.zeros(n_plot_x*n_plot_y)

        # Compute mean residual
        replic_fluxes = np.zeros((n_bands, n_replicated))
        mean_replic_fluxes = np.zeros(n_bands)
        for i, band_name in enumerate((self.filters.data['label'][sor])):
            mean_replic_fluxes[i] = np.mean(replic_data.data['_'+band_name+'_']) * flux_factor
            replic_fluxes[i,:] = replic_data.data['_'+band_name+'_'] * flux_factor

        mean_residual = (mean_replic_fluxes-obs_flux)/obs_flux_err 

        # Compute variance-covariance matrix of residual
        residual_fluxes = (replic_fluxes-obs_flux.reshape(n_bands,
            1).repeat(n_replicated, 1)) / obs_flux_err.reshape(n_bands,
                    1).repeat(n_replicated, 1)

        residual_covar = np.cov(residual_fluxes)
        print("residual_covar: ", residual_covar)

        # Plot the variance-covariance matrix of residuals
        if 'sns' in sys.modules:

            sns.set(style="white")
            labels = list()
            for lab in self.filters.data['label']:
                labels.append(lab.split('_')[-1])
                
            #d = pd.DataFrame(data=np.abs(residual_fluxes.T),
            #        columns=labels)

            # Compute the correlation matrix
            #corr = d.corr()

            # Generate a mask for the upper triangle
            mask = np.zeros_like(corr, dtype=np.bool)
            mask[np.triu_indices_from(mask)] = True

            # Set up the matplotlib figure
            fig, ax = plt.subplots()

            # Generate a custom diverging colormap
            cmap = sns.diverging_palette(220, 10, as_cmap=True)

            # Draw the heatmap with the mask and correct aspect ratio
            sns.heatmap(corr, mask=mask, cmap=cmap, vmax=0.4,
                    square=True, annot=True, fmt=".2f", annot_kws={"size": 10},
                    linewidths=.5, cbar_kws={"shrink": .85}, ax=ax)

            # Rotate by 45 deg the x and y ticks so they do not overlap
            plt.setp( ax.xaxis.get_majorticklabels(), rotation=45,
                    horizontalalignment='right' )

            plt.setp( ax.yaxis.get_majorticklabels(), rotation=45,
                    horizontalalignment='right' )

            name = prepare_plot_saving(str(ID)+'_BEAGLE_replic_data_phot_matrix.pdf')

            fig.savefig(name, dpi=None, facecolor='w', edgecolor='w',
                    orientation='portrait', papertype='a4', format="pdf",
                    transparent=False, bbox_inches="tight", pad_inches=0.1)

            fig.clear()
            plt.close(fig)

        # Select a random set of repliated data
        np.random.seed(seed=12345678)
        replic_data_rows = np.random.choice(n_replicated, size=n_plot_x*n_plot_y)    

        fig, axs = plt.subplots(n_plot_x, n_plot_y, sharex=True, sharey=True)
        fig.subplots_adjust(left=0.08, bottom=0.08, hspace=0, wspace=0)
        fontsize = 8
        axes_linewidth = 0.7

        ix = 0
        iy = 0
        for i, ax in enumerate(np.ravel(axs)):


#            kwargs = {'alpha':0.7}
#            (_, caps, _) = ax.errorbar(wl_eff, 
#                    obs_flux, 
#                    yerr=obs_flux_err, 
#                    ls=' ', 
#                    marker='o', 
#                    markersize=5, 
#                    color='orangered',
#                    markeredgewidth = 0.,
#                    elinewidth=1.0,
#                    capsize=2,
#                    **kwargs)
#
#            for cap in caps:
#                cap.set_color('orangered')
#                cap.set_markeredgewidth(1)
#
            temp_data = replic_data.data[replic_data_rows[i]]
            replic_fluxes = np.array(temp_data[1:]) * flux_factor

            diff_fluxes = (replic_fluxes-obs_flux) / obs_flux_err

            kwargs = {'alpha':0.4}
            unique_markers = np.unique(markers)
            if i != 0:
                for um in unique_markers:
                    mask = markers == um 

                    (_, caps, _) = ax.errorbar(wl_eff[mask], 
                            diff_fluxes[mask], 
                            ls=' ', 
                            marker=um, 
                            markersize=5, 
                            color='black',
                            markeredgewidth = 0.,
                            elinewidth=1.0,
                            capsize=2,
                            **kwargs)

                    for cap in caps:
                        cap.set_color('black')
                        cap.set_markeredgewidth(1)

            if i == 0:

                nXgrid = 1000
                kwargs = {'alpha':0.7}

                delta_wl = wl_eff[1:]-wl_eff[0:-1]
                delta_wl = np.concatenate(([delta_wl[0]], delta_wl))

                for j in range(n_bands):

                    residual = residual_fluxes[j,:]

                    # This function provides you with all the necessary info to draw violin plots
                    kde_pdf, pdf_norm, median_flux, x_plot, y_plot = prepare_violin_plot(residual)

                    w = 0.4 * delta_wl[j] / np.max(y_plot)

                    y_grid = np.full(len(x_plot), wl_eff[j])

                    _lim_y = kde_pdf(median_flux)/pdf_norm * w

                    ax.fill_betweenx(x_plot,
                            y_grid - y_plot*w,
                            y_grid + y_plot*w,
                            **kwargs
                            )

                    ax.plot( [wl_eff[j]-_lim_y, wl_eff[j]+_lim_y],
                            [median_flux, median_flux],
                            color = 'black',
                            linewidth = 0.2
                            )

            #min_flux[i] = np.min(np.array([replic_fluxes-obs_flux_err, obs_flux-obs_flux_err]))
            #max_flux[i] = np.max(np.array([replic_fluxes+obs_flux_err, obs_flux+obs_flux_err]))
            max_abs_flux[i] = np.max(np.abs(diff_fluxes))

        # Determine min and max values of y-axis
        yMax = np.max(max_abs_flux)
        yMin = -yMax
        dY = yMax-yMin
        yMax += dY * 0.1
        yMin -= dY * 0.1

        xMin = wl_eff[0]
        xMax = wl_eff[-1]
        dX = xMax-xMin
        xMax += dX * 0.1
        xMin -= dX * 0.1

        for ax in np.ravel(axs):        
            ax.set_ylim([yMin, yMax])
            ax.set_xlim([xMin, xMax])

            ax.tick_params(which='minor', axis='both',
                            length=2, width=axes_linewidth)

            ax.tick_params(which='major', axis='both',
                            length=3.5, width=axes_linewidth)

            x0, x1 = ax.get_xlim()
            if yMin < 0.: ax.plot( [x0,x1], [0.,0.], color='gray', lw=0.8 )


            for item in ([ax.title, ax.xaxis.label, ax.yaxis.label] +
                         ax.get_xticklabels() + ax.get_yticklabels()):
                item.set_fontsize(fontsize)


            for axis in ['top','bottom','left','right']:
              ax.spines[axis].set_linewidth(axes_linewidth)

            # Set better location of tick marks
            set_plot_ticks(ax, n_x=4, prune_x='both', prune_y='both')

        xlabel = "$\\lambda_\\textnormal{eff} / \\textnormal{\AA}$ (observed-frame)"
        #ylabel = "$f_{\\nu}/\\textnormal{nanoJy}$"
        ylabel = "$\\left(f_{\\nu}^\\textnormal{rep}-f_{\\nu}\\right) / \sigma$"

        fig.text(0.5, 0.02, xlabel, ha='center', fontsize=fontsize+1)
        fig.text(0.03, 0.5, ylabel, va='center', rotation='vertical', fontsize=fontsize+1)

        # Title of the plot is the object ID
        #plt.title(str(ID))

        # Location of printed text
        #x0, x1 = ax.get_xlim()
        #x = x0 + (x1-x0)*0.03
        #y0, y1 = ax.get_ylim()
        #y = y1 - (y1-y0)*0.10

        if print_text:

            # Print the evidence
            try:
                ax.text(x, y, "$\\log(Z)=" + "{:.2f}".format(self.logEvidence) + "$", fontsize=10 )
            except AttributeError:
                print("ciao")

            # Print the average reduced chi-square
            try:
                row = extract_row(self.PPC.data, ID, key=self.key)
                aver_chi_square = row['aver_chi_square']
                y = y1 - (y1-y0)*0.15
                ax.text(x, y, "$\\langle\chi^2\\rangle=" + "{:.2f}".format(aver_chi_square) + "$", fontsize=10 )
            except AttributeError:
                print("`PosteriorPredictiveChecks` not computed/loaded, hence " \
                "<chi^2> for the object `" + str(ID) + "` is not available")

            try:
                row = extract_row(self.PPC.data, ID, key=self.key)
                aver_red_chi_square = row['aver_red_chi_square']
                n_data = row['n_used_bands']
                y = y1 - (y1-y0)*0.20
                ax.text(x, y,
                        "$\\langle\chi^2/(\\textnormal{N}_\\textnormal{data}-1)\\rangle=" \
                        + "{:.2f}".format(aver_red_chi_square) + "\; \
                        (\\textnormal{N}_\\textnormal{data}=" + \
                        "{:d}".format(n_data) + ")" + "$", fontsize=10 )
            except AttributeError:
                print("`PosteriorPredictiveChecks` not computed/loaded, hence " \
                "<chi^2_red> for the object `" + str(ID) + "` is not available")


        #fig.tight_layout()

        name = prepare_plot_saving(plot_name)

        fig.savefig(name, dpi=None, facecolor='w', edgecolor='w',
                orientation='portrait', papertype='a4', format="pdf",
                transparent=False, bbox_inches="tight", pad_inches=0.1)

        plt.close(fig)

        model_hdu.close()
        replic_hdu.close()

##    def plot_residuals(self, residual_file_name=None, residual_plotname=None):
##
##        if not hasattr(self, 'observed_catalogue'):
##            except AttributeError:
##                    "An observed catalogue must be loaded before plotting the
##                    residual"
##
##        if not hasattr(self, 'beagle_summary_catalogue'):
##            except AttributeError:
##                    "A `beagle_summary_catalogue` must be loaded before plotting the
##                    residual"
##
##        self.residual = ResidualPhotometry()
##
##        try:
##            self.residual.load(self.residual_file_name)
##        except:
##            self.residual.compute(self.observed_catalogue,
##                self.beagle_summary_catalogue, self.self.filters.
##                cPickleName=self.residual_file_name)
                





from __future__ import absolute_import
import matplotlib.pyplot as plt

def set_shared_ylabel(a, ylabel, labelpad = 0.01):
    """Set a y label shared by multiple axes
    Parameters
    ----------
    a: list of axes
    ylabel: string
    labelpad: float
        Sets the padding between ticklabels and axis label"""

    f = plt.figure()  
    f.canvas.draw() #sets f.canvas.renderer needed below

    # get the center position for all plots
    top = a[0].get_position().y1
    bottom = a[-1].get_position().y0

    # get the coordinates of the left side of the tick labels 
    x0 = 1
    for at in a:
        at.set_ylabel('') # just to make sure we don't and up with multiple labels
        bboxes, _ = at.yaxis.get_ticklabel_extents(f.canvas.renderer)
        bboxes = bboxes.inverse_transformed(f.transFigure)
        xt = bboxes.x0
        if xt < x0:
            x0 = xt
    tick_label_left = x0

    # set position of label
    a[-1].set_ylabel(ylabel)
    a[-1].yaxis.set_label_coords(tick_label_left-labelpad, (bottom + top)/2, transform=f.transFigure)

def set_shared_xlabel(a, xlabel, labelpad = 0.01):
    """Set a x label shared by multiple axes
    Parameters
    ----------
    a: list of axes
    ylabel: string
    labelpad: float
        Sets the padding between ticklabels and axis label"""

    f = plt.figure()  
    f.canvas.draw() #sets f.canvas.renderer needed below

    # get the center position for all plots
    left = a[0].get_position().x0
    right = a[-1].get_position().x1

    # get the coordinates of the left side of the tick labels 
    y0 = 1
    for at in a:
        at.set_xlabel('') # just to make sure we don't and up with multiple labels
        bboxes, _ = at.xaxis.get_ticklabel_extents(f.canvas.renderer)
        bboxes = bboxes.inverse_transformed(f.transFigure)
        yt = bboxes.y0
        if yt < y0:
            y0 = yt
    tick_label_bottom = y0

    # set position of label
    a[-1].set_xlabel(xlabel)
    a[-1].xaxis.set_label_coords((left + right)/2, tick_label_bottom-labelpad, transform=f.transFigure)



from __future__ import print_function

from __future__ import absolute_import
import time
import os
import logging
from collections import OrderedDict
import json
import numpy as np
from scipy.interpolate import interp1d
from astropy.io import fits
from pathos.multiprocessing import ProcessingPool 
from natsort import index_natsorted, order_by_index

from .beagle_utils import prepare_data_saving, BeagleDirectories, getPathForData, data_exists,\
    ID_COLUMN_LENGTH
from .significant_digits import to_precision
import six
from six.moves import range

def get1DInterval(param_values, probability, levels):

    """ 
    Compute several quantities from a 1D probability density function

    Parameters
    ----------
    param_values : numpy array
        Contains the values of the parameter.

    probability : numpy array 
        Contains the probability associated with each value of the
        parameter, hence must have same dimension as 'param_values'.

    levels : numpy array or list containing float
        Contains the (percentage) levels used to compute the credible
        regions, e.g. levels=[68.,95.] will compute 68 % and 95 % (central)
        credible regions

    Returns
    -------
    mean : float
        Mean of the parameter, computed as
        sum(probability*param_values)/sum(probability)
    median : float
        Median of the parameter, computed from the cumulative integral of
        the PDF
    interval : list of float
        2-dimensional list containing the lower and upper value of the
        parameter corresponding to the different `levels`

    """

    sort_ = np.argsort(param_values)

    # ******************************************************************
    # Here you must simply use `cumsum`, and not `cumtrapz` as in
    # beagle_utils.prepare_violin_plot, since the output of MultiNest are a set
    # of weights (which sum up to 1) associated to each set of parameters (the
    # `p_j` of equation 9 of Feroz+2009), and not a probability density (as the
    # MultiNest README would suggest).
    # ******************************************************************
    cumul_pdf = np.cumsum(probability[sort_])
    cumul_pdf /= cumul_pdf[len(cumul_pdf)-1]

    # Get the interpolant of the cumulative probability
    f_interp = interp1d(cumul_pdf, param_values[sort_])

    # You shoud integrate rather than summing here
    mean = np.sum(probability * param_values) / np.sum(probability)

    median = f_interp(0.5)

    interval = list()
    for lev in levels:

        low, high = f_interp([0.5*(1.-lev/100.), 1.-0.5*(1.-lev/100.)])
        interval.append([low,high])

    return mean, median, interval


class BeagleSummaryCatalogue(object):

    def __init__(self, 
            file_name=None, 
            credible_intervals=None,
            config_file=None,
            hdu_col=None,
            n_proc=1):

        if file_name is not None:
            self.file_name  = file_name
        else:
            self.file_name = "BEAGLE_summary_catalogue.fits" 

        self.hdu_col = hdu_col

        if config_file is not None:
            if os.path.dirname(config_file) is None:
                _config_file  = os.path.join(BeagleDirectories.results_dir, 
                        config_file)
            else:
                _config_file = config_file

            with open(_config_file) as f:    
                self.hdu_col = json.load(f, object_pairs_hook=OrderedDict)

        else:
            _config_file = os.path.join(BeagleDirectories.results_dir, "summary_config.json")
            if os.path.isfile(_config_file):
                with open(_config_file) as f:    
                    self.hdu_col = json.load(f, object_pairs_hook=OrderedDict)

        self.exclude_columns = ['probability', 'ln_likelihood', 'chi_square', 'n_data'] 

        if self.hdu_col is None:
            self.hdu_col = list()
            self.hdu_col.append({'name':'POSTERIOR PDF'})

        self.credible_intervals = credible_intervals

        self.n_proc = n_proc

    def exists(self):

        return data_exists(self.file_name)

    def load(self):
        """ 
        Load a 'BEAGLE summary catalogue'

        Parameters
        ----------
        file_name : str
            Name of the file containing the catalogue.
        """

        logging.info("Loading the `BeagleSummaryCatalogue` file: " + self.file_name)

        name = getPathForData(self.file_name)
        self.hdulist = fits.open(name)

    def compute_single(self, file, hdu_col):
        """ 
        """ 

        # Compute the required quantities
        hdulist = fits.open(os.path.join(BeagleDirectories.results_dir, file))
        end = file.find('_' + BeagleDirectories.suffix)

        # Extract the object ID from the file_name
        #ID = np.int(np.float(os.path.basename(file[0:end])))
        ID = os.path.basename(file[0:end])

        probability = hdulist['posterior pdf'].data['probability']
        data = OrderedDict()

        for hdu in hdu_col:
            hdu_name = hdu['name']
            if 'columns' in hdu:
                columnNames = hdu['columns']
            else:
                columnNames = hdulist[hdu_name].columns.names

            for col_name in columnNames:
                data['ID'] = ID
                par_values = hdulist[hdu_name].data[col_name]

                mean, median, interval = get1DInterval(par_values, probability, self.credible_intervals)

                data[col_name+'_mean'] = mean
                data[col_name+'_median'] = median

                for j, lev in enumerate(self.credible_intervals):
                    levName = col_name + '_' + "{:.2f}".format(lev)
                    data[levName] = interval[j]

        hdulist.close()

        return data

    def compute(self, file_list, overwrite=False):
        """ 
        """ 

        # You consider the first file in the list and use as a "mold" to create
        # the structure (binary tables and their columns) of the output FITS file
        firstfile = os.path.join(BeagleDirectories.results_dir, file_list[0])
        hdulist = fits.open(firstfile)

        n_objects = len(file_list)

        # Initialize a new (empty) primary HDU for your output FITS file
        self.hdulist = fits.HDUList(fits.PrimaryHDU())
    
        # Now you cycle over all extension and columns that you want to put in
        # the summary catalogue
        for hdu in self.hdu_col:
            new_columns = list()

            # Pick the extension name
            hdu_name = hdu['name']

            # The first column of each output extension contains the object ID
            #new_columns.append(fits.Column(name='ID', format='K'))
            new_columns.append(fits.Column(name='ID', format=str(ID_COLUMN_LENGTH)+'A'))

            # You just consider the columns defined in the structure
            if 'columns' in hdu:
                columnNames = hdu['columns']
            # While by default you take all columns in that extensions
            else:
                columnNames = [name for name in hdulist[hdu_name].columns.names if name not in self.exclude_columns]

            # For each column, you add a '_mean', '_median' and confidence
            # intervals columns, taking the appropriate units from the FITS
            # file that you are using as a mold
            for col_index in range(len(hdulist[hdu_name].columns)):

                col_ = hdulist[hdu_name].columns[col_index]

                if col_.name not in columnNames:
                    continue
    
                new_columns.append(fits.Column(name=col_.name+'_mean',
                    format=col_.format, unit=col_.unit))

                new_columns.append(fits.Column(name=col_.name+'_median',
                    format=col_.format, unit=col_.unit))

                for lev in self.credible_intervals:
                    new_columns.append(fits.Column(name=col_.name + '_' +
                        "{:.2f}".format(lev), format='2'+col_.format[-1],
                        unit=col_.unit))

            # Create the "column definition"
            cols_ = fits.ColDefs(new_columns)

            # Create the actual binary table, with the correct number of rows
            # to accomodate all objects
            nrows = len(file_list)
            new_hdu = fits.BinTableHDU.from_columns(cols_, nrows=nrows)
            new_hdu.name = hdu_name

            # And finally append the newly created bunary table to the hdulist
            # that will be printed to the ouput FITS file
            self.hdulist.append(new_hdu)

        hdulist.close()

        #start_time = time.time()
        # Now you can go through each file, and compute the required quantities
        if self.n_proc > 1:
            pool = ProcessingPool(nodes=self.n_proc)
            data = pool.map(self.compute_single, 
                    file_list,
                    (self.hdu_col,)*len(file_list))
        else:
            data = list()
            for i, file in enumerate(file_list):
                d = self.compute_single(file, self.hdu_col)
                data.append(d)

        #print("--- %s seconds ---" % (time.time() - start_time))

        # Natural sort IDs
        IDs = [data[i]['ID'] for i in range(len(data))]
        index = index_natsorted(IDs)

        for i, file in enumerate(file_list):
            idx = index[i]
            for hdu in self.hdu_col:
                hdu_name = hdu['name']
                if 'columns' in hdu:
                    columnNames = hdu['columns']
                else:
                    columnNames = [name for name in hdulist[hdu_name].columns.names if name not in self.exclude_columns]

                for col_name in columnNames:
                    self.hdulist[hdu_name].data['ID'][i] = data[idx]['ID']
                    self.hdulist[hdu_name].data[col_name+'_mean'][i] = data[idx][col_name+'_mean']
                    self.hdulist[hdu_name].data[col_name+'_median'][i] = data[idx][col_name+'_median']
                    for j, lev in enumerate(self.credible_intervals):
                        levName = col_name + '_' + "{:.2f}".format(lev)
                        self.hdulist[hdu_name].data[levName][i] = data[idx][levName]

        name = prepare_data_saving(self.file_name)
        self.hdulist.writeto(name, clobber=overwrite)

    def extract_MAP_solution(self, file_list, overwrite=False):
        """ 
        """ 

        # Now you can go through each file, and extract the MAP solution
        for i, file in enumerate(file_list):

            # Open the original BEAGLE FITS file
            hdulist = fits.open(os.path.join(BeagleDirectories.results_dir, file))

            # Get the posterior probability
            post = hdulist['posterior pdf'].data['probability']

            # Maximum of the posterior PDF, i.e. you select the template
            # corresponding to the mode of the posterior distributions
            MAP_indx = np.argmax(post)

            # Now write in a separate FITS file the pararmeters corresponding to the MAP row
            new_hdulist = fits.HDUList(fits.PrimaryHDU())

            for hdu in hdulist:

                if hdu.data is None:
                    continue

                if hdu.is_image:
                    new_hdu = fits.PrimaryHDU()
                    new_hdu.name = hdu.name
                    new_hdu.data = hdu.data[MAP_indx,:]
                else:
                    new_hdu = fits.BinTableHDU.from_columns(hdu.columns, nrows=1)
                    new_hdu.name = hdu.name
                    if 'sed wl' in hdu.name.lower() or 'sed mask' in hdu.name.lower():
                        new_hdu.data = hdu.data
                    else:
                        new_hdu.data[0] = hdu.data[MAP_indx]

                new_hdulist.append(new_hdu)

            # Extract the object ID from the file_name
            end = file.find('_' + BeagleDirectories.suffix)
            ID = os.path.basename(file[0:end])

            file_name = ID + '_BEAGLE_MAP.fits.gz'
            name = prepare_data_saving(file_name)
            new_hdulist.writeto(name, clobber=overwrite)

            new_hdulist.close()
            hdulist.close()

    def make_latex_table(self, param_names, 
            IDs=None,
            summary_statistics='median',
            average_errors=True,
            significant_digits=1):

        # Check if a list of params has been passed, or a JSON file
        if param_names[0].lower().endswith("json"):
            with open(param_names[0]) as f:    
                param_config = json.load(f, object_pairs_hook=OrderedDict)
        else:
            param_config = OrderedDict()
            for name in param_names:
                param_config[name] = {}

        n = significant_digits
        
        n_rows = len(self.hdulist[1].data['ID'])

        if IDs is None:
            IDs = self.hdulist[1].data['ID']

        for ID in IDs:
            print("\n " + ID, end='') 
            row = np.arange(n_rows)[self.hdulist[1].data['ID'] == ID][0]
            for param, value in six.iteritems(param_config):
                is_log = False
                if "log" in value:
                    is_log = value["log"]

                factor = 1.0
                if "factor" in value:
                    factor = value["factor"]

                colName = param + '_' + summary_statistics
                for hdu in self.hdulist:
                    if hasattr(hdu, 'columns'):
                        if colName in hdu.columns.names:
                            sum_stat = hdu.data[colName][row]
                            #print "n_ ", n_
                            for lev in self.credible_intervals:
                                errColName = param + '_' + "{:.2f}".format(lev)
                                error_low, error_up = hdu.data[errColName][row]

                                if average_errors:
                                    err = 0.5 * (error_up-error_low)
                                    err /= factor ; sum_stat /= factor
                                    if is_log:
                                        err = 0.434*err/sum_stat
                                        sum_stat = np.log10(sum_stat)
                                    n_ = int(np.ceil(np.log10(abs(sum_stat)))) + 1
                                    print(' & $' + to_precision(sum_stat, n+n_) + '\\pm' + to_precision(err, n) + '$', end='')
                                else:
                                    error_low = sum_stat - error_low
                                    error_up = error_up - sum_stat
                                    error_low /= factor ; error_up /= factor ; sum_stat /= factor
                                    if is_log:
                                        error_low = 0.434*error_low/sum_stat
                                        error_up = 0.434*error_up/sum_stat
                                        sum_stat = np.log10(sum_stat)
                                    n_ = int(np.ceil(np.log10(abs(sum_stat)))) + 1
                                    print(' & $' + to_precision(sum_stat, n+n_) + '_{-' + to_precision(error_low, n) \
                                            + '}^{+' + to_precision(error_up, n) + '}$', end='')
                            break
            print(' \\\\ ', end='')

                        



from __future__ import absolute_import
from __future__ import print_function
import os
from scipy import stats
import numpy as np
import matplotlib.pyplot as plt
from astropy.io import fits
from astropy.table import Table, Column

import sys
import pyp_beagle.dependencies.WeightedKDE 
from pyp_beagle.dependencies.walker_random_sampling import WalkerRandomSampling

from .beagle_utils import prepare_data_saving, prepare_plot_saving, \
    BeagleDirectories, set_plot_ticks
from six.moves import range

# 1 jy = 10^-23 erg s^-1 cm^-2 hz^-1
jy = 1.E-23 

class PosteriorPredictiveChecks(object):

    def chi_square(self, y, E_y, sig_y):
        """ 
        An example of a "discrepancy function", the chi-square

        Parameters
        ----------
        y : float
            "Observed" value

        E_y : float
            "Expected" (theoretical) value

        sig_y: float
            Standard deviation of the "observed" values

        Returns 
        ------
        float    

        Notes
        -----
        Negative values of sig_y produce a mask, i.e. those values are not
        considered in the chi-square computation.
        For further details on "discrepancy functions" see Gelman, Meng & Stern (1996).
        """
    
        if y.ndim == 1:
            loc = np.where(sig_y > 0.)[0]
            return np.sum((y[loc]-E_y[loc])**2/sig_y[loc]**2)
        elif y.ndim == 2:
            my = np.ma.masked_array(y, mask=(sig_y <= 0.))
            mE_y = np.ma.masked_array(E_y, mask=(sig_y <= 0.))
            msig_y = np.ma.masked_array(sig_y, mask=(sig_y <= 0.))
            return np.ma.sum((my-mE_y)**2/msig_y**2, axis=0)


    def load(self, file_name):
        """ 
        Load a file containing the posterior predictive checks

        Parameters
        ----------
        file_name : str
            Name of the file.
        """

        name = os.path.join(BeagleDirectories.results_dir,
                BeagleDirectories.pypbeagle_data, file_name)

        my_table = Table.read(name)
    
        self.data = my_table

    def compute_replicated(self, observed_catalogue, filters, ID,
            n_replicated=2000, seed=1234):

            strID = str(ID)
            file = os.path.join(BeagleDirectories.results_dir,
                    strID + '_' + BeagleDirectories.suffix + '.fits.gz')
            
            # Name of the output file containing the replicated data
            out_name = strID + "_BEAGLE_replic_data.fits.gz"
            out_name = prepare_data_saving(out_name)

            # Check if the file containing the replicated data already exists, in which case just read it!
            if os.path.isfile(out_name):

                # Read the model fluxes from the BEAGLE output file
                hdulist = fits.open(file)
                beagle_data = hdulist['MARGINAL PHOTOMETRY'].data

                n_samples = len(hdulist['MARGINAL PHOTOMETRY'].data.field(0))
                cols = hdulist['MARGINAL PHOTOMETRY'].columns

                model_flux = np.zeros((filters.n_bands, n_samples), np.float32)

                for j in range(filters.n_bands):
                    name = '_' + filters.data['label'][j] + '_'
                    model_flux[j,:] = beagle_data[name] / jy

                # Close the BEAGLE output file and open the file containing replicated data
                hdulist.close()
                hdulist = fits.open(out_name)

                # Get the row indices, indicating which rows from the output
                # BEAGLE file were used in the creation of the replicated data
                replic_data_rows = hdulist[1].data['row_index']

                noiseless_flux = model_flux[:, replic_data_rows]

                replic_flux = np.zeros((filters.n_bands, len(replic_data_rows)), np.float32)
                for col_name in hdulist[1].columns.names:
                    if col_name in cols.names:
                        replic_flux[j,:] = hdulist[1].data[col_name]

                return replic_flux, noiseless_flux, model_flux, n_data

            if os.path.isfile(file):

                # Open the FITS file containing BEAGLE results for the current object
                hdulist = fits.open(file)
                beagle_data = hdulist['MARGINAL PHOTOMETRY'].data

                # Load the posterior probability and create array of row indices
                probability = hdulist['POSTERIOR PDF'].data['probability']
                n_samples = len(probability)
                row_indices = np.arange(n_samples)

                # Now, draw the weighted samples with replacement
                wrand = WalkerRandomSampling(probability, keys=row_indices, rand_seed=seed)
                replic_data_rows = wrand.random(n_replicated)

                obs_flux = np.zeros(filters.n_bands, np.float32)
                obs_flux_err = np.zeros(filters.n_bands, np.float32)
                model_flux = np.zeros((filters.n_bands, n_samples), np.float32)

                # The replicated data are just the fluxes predicted by your
                # model, drawn from the posterior probability distribution
                # accordingly to their probability, with the effect of
                # observatironal noise added
                noiseless_flux = np.zeros((filters.n_bands, n_replicated), np.float32)
                replic_flux = np.zeros((filters.n_bands, n_replicated), np.float32)

                n_data = 0

                obs_flux, obs_flux_err = observed_catalogue.extract_fluxes(filters, ID)
                n_data = np.count_nonzero(obs_flux_err > 0.)

                for j in range(filters.n_bands):

                    # model flux
                    name = '_' + filters.data['label'][j] + '_'
                    model_flux[j,:] = beagle_data[name] / jy

                # You save in this array the noise-less flux predicted by the model        
                noiseless_flux = model_flux[:, replic_data_rows]

                for j in range(filters.n_bands):
                    # Here you add the observational noise to the
                    # noise-less fluxes predicted by the model, obtaining
                    # the actual replicated data
                    if obs_flux_err[j] > 0.:
                        replic_flux[j,:] = noiseless_flux[j, :] +   \
                        np.random.normal(scale=obs_flux_err[j], size=n_replicated)
                    else:    
                        replic_flux[j,:] = -99.99999

                # Write the replicated data to an output FITS file
                # Create the new FITS file
                new_hdu = fits.HDUList(fits.PrimaryHDU())

                # Add column allowing you to match each row of the replicated
                # data to the row of noise-less fluxes predicted by your model,
                # i.e. those in the "MARGINAL PHOTOMETRY" extension of the
                # BEAGLE output FITS file 
                # NB: the column indexing start with 1 !!
                ID_col = fits.Column(name='row_index', format='I')

                # Copy the columns defined in the "MARGINAL PHOTOMETRY"
                # extension to the new FITS file
                cols = hdulist['MARGINAL PHOTOMETRY'].columns

                new_hdu.append(fits.BinTableHDU.from_columns(ID_col + cols, nrows=n_replicated, fill=True))

                # Fill with the replicated data fluxes
                j = 0
                for col_name in new_hdu[1].columns.names:
                    if col_name in cols.names:
                        new_hdu[1].data[col_name] = replic_flux[j,:]
                        j += 1

                # Fill with the row indices
                new_hdu[1].data['row_index'] = replic_data_rows

                # Save the file
                new_hdu.writeto(out_name)

            hdulist.close()

            return replic_flux, noiseless_flux, model_flux, n_data

    def compute(self, observed_catalogue, filters, discrepancy=None, 
            n_replicated=2000, file_name=None):
        """ 
        Compute  posterior predictive checks quantities.

        Parameters
        ----------
        observed_catalogue : `beagle_photometry.PhotometricCatalogue`
            Class containing an observed photometric catalogue.

        filters : `beagle_filters.PhotometricFilters`
            Class containing a set of photometric filters.

        discrepancy : function, optional
            The discrepancy function used in the posterior predicitve check.

        n_replicated: int, optional
            The number of replicated data to draw.

        file_name : str, optional
            Name of the output catalogue, wuthout including the direcory tree.
            It will be saved into the RESULTS_DIR/pypbeagle_DATA folder (which
            will be created if not present).
        """

        if file_name is None:
            file_name = "PPC.fits"

        if discrepancy is None:
            discrepancy = self.chi_square

        # Copy from the catalogue the column containing the object IDs
        objID = Column(data=observed_catalogue.data['ID'], name='ID', dtype=np.int32) 

        n_obj = len(observed_catalogue.data['ID'])
        
        # Defines columns containing the number of photometric bands actually
        # used in the BEAGLE run, for a given object,
        n_used_bands = Column(name='n_used_bands', dtype=np.int32, length=n_obj)

        deg_of_freedom = Column(name='dof', dtype=np.int32, length=n_obj)

        # the average chi-square, and average chi-square/n_used_bands
        # Compute the "average chi square", a measure of the predicitve accuacy
        # of the model (e.g. see Section 6.5 of "Bayesian Data Analysis", by
        # Gelman, Carlin, Stern and Rubin )
        aver_chi_square = Column(name='aver_chi_square', dtype=np.float32,
                length=n_obj)

        # Int from 0 to x of chi^2(x) with N-1 degreed of freedom (see Johnson,
        # V. E. (2004). A Bayesian chi2 Test for Goodness-of-Fit on JSTOR.
        # Annals of Statistics for an explanation of why the average chi^2 has
        # N-1 and not N-k-1 degrees of freedom)

        left_cumul_probability = Column(name='left_cumul_probability', dtype=np.float32,
                length=n_obj)

        # Int from x to +infty of chi^2(x)
        right_cumul_probability = Column(name='right_cumul_probability', dtype=np.float32,
                length=n_obj)

        aver_red_chi_square = Column(name='aver_red_chi_square',
                dtype=np.float32, length=n_obj)

        p_value = Column(name='p_value',
                dtype=np.float32, length=n_obj)

        my_cols = [objID, n_used_bands, deg_of_freedom, aver_chi_square,
                aver_red_chi_square, left_cumul_probability,
                right_cumul_probability, p_value]

        my_table = Table(my_cols)

        model_flux = np.zeros(filters.n_bands, np.float32)


        for i in range(n_obj):

            ID = objID[i]
            strID = str(objID[i])
            file = os.path.join(BeagleDirectories.results_dir,
                    strID + '_' + BeagleDirectories.suffix + '.fits.gz')

            if os.path.isfile(file):

                print("")
                print("HERE")
                obs_flux, obs_flux_err = observed_catalogue.extract_fluxes(filters, ID)

                replic_flux, noiseless_flux, model_flux, n_data = self.compute_replicated(observed_catalogue, filters, ID)

                hdulist = fits.open(file)
                probability = hdulist['POSTERIOR PDF'].data['probability']

                n_samples = model_flux.shape[1]
                n_replicated = replic_flux.shape[1]

                # Extend the arrays containing the observed flux and its
                # error to match the shape of the replicated data array
                ext_obs_flux = obs_flux.reshape(filters.n_bands, 1).repeat(n_replicated, 1)
                ext_obs_flux_err = obs_flux_err.reshape(filters.n_bands, 1).repeat(n_replicated, 1)

                # Compute the "discrepancy" for the actual data, and for the replicated data
                discrepancy_data = discrepancy(ext_obs_flux, noiseless_flux, ext_obs_flux_err)
                discrepancy_repl_data = discrepancy(replic_flux, noiseless_flux, ext_obs_flux_err)

                # The p-value is just the fraction of objects for which the
                # discrepancy for the replicated data is larger then that for
                # the actual data! 
                my_table['p_value'][i] = 1. * np.count_nonzero((discrepancy_repl_data > discrepancy_data)) / n_replicated

                dof = n_data
                my_table['n_used_bands'][i] = n_data
                my_table['dof'][i] = dof

                # Here you cosider all samples in the posterior
                ext_obs_flux = obs_flux.reshape(filters.n_bands, 1).repeat(n_samples, 1)
                ext_obs_flux_err = obs_flux_err.reshape(filters.n_bands, 1).repeat(n_samples, 1)
                av_chi_square = np.sum(probability*self.chi_square(ext_obs_flux, model_flux, ext_obs_flux_err)) / np.sum(probability)
                my_table['aver_chi_square'][i] = av_chi_square
                my_table['aver_red_chi_square'][i] = av_chi_square / dof

                cdf = stats.chi2.cdf(av_chi_square, dof)
                my_table['left_cumul_probability'][i] = cdf
                my_table['right_cumul_probability'][i] = 1.-cdf

                hdulist.close()

        self.columns = my_cols
        self.data = my_table

        name = prepare_data_saving(file_name)
        my_table.write(name)

    def plot_chi2(self, 
            plot_name="BEAGLE_average_chi_square.pdf"):
        """ 
        Plots the distribution (histogram) of the average chi-square.

        Parameters
        ----------
        plot_name: str, optional
            File name of the output plot.
        """ 

        xdata = self.data['aver_chi_square']
        n_data = len(xdata)
        min_x = 0.
        max_x = 50.

        fig = plt.figure()
        ax = fig.add_subplot(1, 1, 1)

        ax.set_ylabel("Number of galaxies")
        ax.set_xlabel("$\\langle \chi^2 \\rangle$")

        ax.set_xlim((min_x, max_x))

        # Set the correct number of major and mnor tick marks
        set_plot_ticks(ax, prune_y='lower')

        # Plot the histogram of the average chi-square
        kwargs = {'alpha':0.7, 'linewidth':0.5}
        n, bins, patches = ax.hist(xdata, 
                bins=50, 
                range=(min_x, max_x),
                color='gray',
                **kwargs)

        # Integral
        dx = bins[1:]-bins[0:-1]
        norm = np.sum(n*dx)

        # Now, compute the theoretical distribution, i.e. the sum of different
        # chi-square distribution for the different degrees of freedom
        mask = np.zeros(n_data, dtype=bool)
        mask[self.data['dof'] > 0] = True

        min_dof = np.amin(self.data['dof'][mask])
        max_dof = np.amax(self.data['dof'][mask])
                
        dof_range = np.arange(min_dof, max_dof+1)
        frac_data = np.zeros(len(dof_range), dtype=np.float32)

        for i in range(len(dof_range)):
            dof = dof_range[i]
            loc = np.where(self.data['dof'] == dof)[0]
            frac_data[i] = 1.*len(loc)/n_data

        frac_data /= np.sum(frac_data)

        xdata = np.linspace(min_x, max_x, 1000)
        chi_distr = np.zeros(len(xdata), dtype=np.float32)
        for i in range(len(dof_range)):
            dof = dof_range[i]
            chi_distr += frac_data[i] * stats.chi2.pdf(xdata, dof)

        chi_distr *= norm
        ax.plot(xdata, 
                chi_distr,
                color='black',
                linestyle='--')

        name = prepare_plot_saving(plot_name)
        fig.savefig(name, dpi=None, facecolor='w', edgecolor='w',
                orientation='portrait', papertype='a4', format="pdf",
                transparent=False, bbox_inches="tight", pad_inches=0.1)

        plt.close(fig)

    def plot_p_value(self, 
            plot_name="BEAGLE_p_value.pdf", broken_axis=False):
        """ 
        Plots the distribution (histogram) of the p-value.

        Parameters
        ----------
        plot_name : str, optional
            File name of the output plot.

        broken_axis : bool, optionak
            If True, then the y-axis is broken to allow a larger dynamic range
            without loosing details.
        """ 

        xdata = self.data['p_value']
        n_data = len(xdata)
        min_x = 0.
        max_x = 1.

        fig = plt.figure()
        
        if broken_axis:
            fig, axs = plt.subplots(2, 1, sharex=True)
            fig.subplots_adjust(left=0.13, bottom=0.10)
        else:
            fig, axs = plt.subplots(1, 1)
            axs = (axs,)

        ylabel = "Number of galaxies"
        xlabel = "$p$-value"

        fig.text(0.5, 0.02, xlabel, ha='center')
        fig.text(0.03, 0.5, ylabel, va='center', rotation='vertical')

        # Plot the histogram of the average chi-square
        kwargs = {'alpha':0.7, 'linewidth':0.5}
        for ax in axs:
            n, bins, patches = ax.hist(xdata, 
                    bins=50, 
                    range=(min_x, max_x),
                    color='gray',
                    **kwargs)

            ax.set_xlim((min_x, max_x))


        if broken_axis:
            # Set the correct number of major and mnor tick marks
            set_plot_ticks(axs[0], n_x=4, n_y=3, prune_y='lower')
            set_plot_ticks(axs[1], n_x=4, n_y=3, prune_y='both')

            max_y = np.max(n[1:])
            axs[1].set_ylim((0, max_y*1.12))
            axs[0].set_ylim((n[0]*0.8, n[0]*1.04))

            # hide the spines between ax and ax2
            axs[0].spines['bottom'].set_visible(False)
            axs[1].spines['top'].set_visible(False)
            axs[0].xaxis.tick_top()
            axs[0].tick_params(labeltop='off')  # don't put tick labels at the top
            axs[1].xaxis.tick_bottom()

            d = .015  # how big to make the diagonal lines in axes coordinates
            # arguments to pass plot, just so we don't keep repeating them
            kwargs = dict(transform=axs[0].transAxes, color='k', clip_on=False)
            axs[0].plot((-d, +d), (-d, +d), **kwargs)        # top-left diagonal
            axs[0].plot((1 - d, 1 + d), (-d, +d), **kwargs)  # top-right diagonal

            kwargs.update(transform=axs[1].transAxes)  # switch to the bottom axes
            axs[1].plot((-d, +d), (1 - d, 1 + d), **kwargs)  # bottom-left diagonal
            axs[1].plot((1 - d, 1 + d), (1 - d, 1 + d), **kwargs)  # bottom-right diagonal
        else:
            set_plot_ticks(axs[0], n_x=4, n_y=4, prune_y='lower')
            axs[0].set_ylim((0, np.max(n)*1.1))



       # y0, y1 = ax.get_ylim()
        levels = (0.01, 0.05)
        for lev in levels:
            l = lev
            frac = 1.*len(np.where(xdata <= l)[0])/n_data
            print("Fraction of galaxies with p-value < " + "{:.2f}".format(lev) + " = {:.2f}".format(frac))
           # ax.plot((l, l),
           #         (y0, y1),
           #         color='black',
           #         linestyle='--')

        name = prepare_plot_saving(plot_name)
        fig.savefig(name, dpi=None, facecolor='w', edgecolor='w',
                orientation='portrait', papertype='a4', format="pdf",
                transparent=False, bbox_inches="tight", pad_inches=0.1)

        plt.close(fig)




from __future__ import absolute_import
import numpy as np
import numpy.ma as ma 
import matplotlib.pyplot as plt
from six.moves import zip


def fill_between_steps(ax, x, y1, y2=0, step_where='pre', **kwargs):
    ''' fill between a step plot and 

    Parameters
    ----------
    ax : Axes
       The axes to draw to

    x : array-like
        Array/vector of index values.

    y1 : array-like or float
        Array/vector of values to be filled under.
    y2 : array-Like or float, optional
        Array/vector or bottom values for filled area. Default is 0.

    step_where : {'pre', 'post', 'mid'}
        where the step happens, same meanings as for `step`

    **kwargs will be passed to the matplotlib fill_between() function.

    Returns
    -------
    ret : PolyCollection
       The added artist

    '''
    if step_where not in {'pre', 'post', 'mid'}:
        raise ValueError("where must be one of {{'pre', 'post', 'mid'}} "
                         "You passed in {wh}".format(wh=step_where))

    # make sure y values are up-converted to arrays 
    if np.isscalar(y1):
        y1 = np.ones_like(x) * y1

    if np.isscalar(y2):
        y2 = np.ones_like(x) * y2

    # temporary array for up-converting the values to step corners
    # 3 x 2N - 1 array 

    vertices = np.vstack((x, y1, y2))

    # this logic is lifted from lines.py
    # this should probably be centralized someplace
    if step_where == 'pre':
        steps = ma.zeros((3, 2 * len(x) - 1), np.float)
        steps[0, 0::2], steps[0, 1::2] = vertices[0, :], vertices[0, :-1]
        steps[1:, 0::2], steps[1:, 1:-1:2] = vertices[1:, :], vertices[1:, 1:]

    elif step_where == 'post':
        steps = ma.zeros((3, 2 * len(x) - 1), np.float)
        steps[0, ::2], steps[0, 1:-1:2] = vertices[0, :], vertices[0, 1:]
        steps[1:, 0::2], steps[1:, 1::2] = vertices[1:, :], vertices[1:, :-1]

    elif step_where == 'mid':
        steps = ma.zeros((3, 2 * len(x)), np.float)
        steps[0, 1:-1:2] = 0.5 * (vertices[0, :-1] + vertices[0, 1:])
        steps[0, 2::2] = 0.5 * (vertices[0, :-1] + vertices[0, 1:])
        steps[0, 0] = vertices[0, 0]
        steps[0, -1] = vertices[0, -1]
        steps[1:, 0::2], steps[1:, 1::2] = vertices[1:, :], vertices[1:, :]
    else:
        raise RuntimeError("should never hit end of if-elif block for validated input")

    # un-pack
    xx, yy1, yy2 = steps

    # now to the plotting part:
    return ax.fill_between(xx, yy1, y2=yy2, **kwargs)

fig, ax_list = plt. subplots(3, 1)
x = y = np.arange(5)

for ax, where in zip(ax_list, ['pre', 'post', 'mid']):
    ax.step(x, y, where=where, color='r', zorder=5, lw=5)
    fill_between_steps(ax, x, y, 0, step_where=where)
